from miscellaneous import Miscellaneous

stack_f = 7
stack_p = None
box_size = 48
num_points = 401

string = Miscellaneous.compose_suffix(stack_f=stack_f, stack_p=stack_p, box_size=box_size,
                                      num_points=num_points)

print ("length of string: " + str(len(string)) + ", content: " + string)



# -*- coding: UTF-8 -*-

from __future__ import print_function

import numpy as np
from time import time

# Let's take the randomness out of random numbers (for reproducibility)
np.random.seed(0)

size = 4096
A, B = np.random.random((size, size)), np.random.random((size, size))
C, D = np.random.random((size * 128,)), np.random.random((size * 128,))
E = np.random.random((int(size / 2), int(size / 4)))
F = np.random.random((int(size / 2), int(size / 2)))
F = np.dot(F, F.T)
G = np.random.random((int(size / 2), int(size / 2)))

# Matrix multiplication
N = 20
t = time()
for i in range(N):
    np.dot(A, B)
delta = time() - t
print('Dotted two %dx%d matrices in %0.2f s.' % (size, size, delta / N))
del A, B

# Vector multiplication
N = 5000
t = time()
for i in range(N):
    np.dot(C, D)
delta = time() - t
print('Dotted two vectors of length %d in %0.2f ms.' % (size * 128, 1e3 * delta / N))
del C, D

# Singular Value Decomposition (SVD)
N = 3
t = time()
for i in range(N):
    np.linalg.svd(E, full_matrices = False)
delta = time() - t
print("SVD of a %dx%d matrix in %0.2f s." % (size / 2, size / 4, delta / N))
del E

# Cholesky Decomposition
N = 3
t = time()
for i in range(N):
    np.linalg.cholesky(F)
delta = time() - t
print("Cholesky decomposition of a %dx%d matrix in %0.2f s." % (size / 2, size / 2, delta / N))

# Eigendecomposition
t = time()
for i in range(N):
    np.linalg.eig(G)
delta = time() - t
print("Eigendecomposition of a %dx%d matrix in %0.2f s." % (size / 2, size / 2, delta / N))


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'main_gui.ui'
#
# Created by: PyQt5 UI code generator 5.15.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(1098, 824)
        MainWindow.setLocale(QtCore.QLocale(QtCore.QLocale.English, QtCore.QLocale.UnitedStates))
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.frame = QtWidgets.QFrame(self.centralwidget)
        self.frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame.setObjectName("frame")
        self.verticalLayout_2.addWidget(self.frame)
        self.tool_line = QtWidgets.QHBoxLayout()
        self.tool_line.setObjectName("tool_line")
        self.control_area = QtWidgets.QFrame(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.control_area.sizePolicy().hasHeightForWidth())
        self.control_area.setSizePolicy(sizePolicy)
        self.control_area.setFrameShape(QtWidgets.QFrame.Box)
        self.control_area.setFrameShadow(QtWidgets.QFrame.Raised)
        self.control_area.setObjectName("control_area")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout(self.control_area)
        self.horizontalLayout_3.setContentsMargins(-1, -1, 9, -1)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.comboBox_back = QtWidgets.QComboBox(self.control_area)
        self.comboBox_back.setMinimumSize(QtCore.QSize(165, 0))
        self.comboBox_back.setEditable(False)
        self.comboBox_back.setCurrentText("")
        self.comboBox_back.setMaxVisibleItems(14)
        self.comboBox_back.setObjectName("comboBox_back")
        self.horizontalLayout_3.addWidget(self.comboBox_back)
        self.pushButton_pause = QtWidgets.QPushButton(self.control_area)
        self.pushButton_pause.setObjectName("pushButton_pause")
        self.horizontalLayout_3.addWidget(self.pushButton_pause)
        self.pushButton_start = QtWidgets.QPushButton(self.control_area)
        icon = QtGui.QIcon.fromTheme("SP_MediaPlay")
        self.pushButton_start.setIcon(icon)
        self.pushButton_start.setObjectName("pushButton_start")
        self.horizontalLayout_3.addWidget(self.pushButton_start)
        self.pushButton_next_job = QtWidgets.QPushButton(self.control_area)
        self.pushButton_next_job.setObjectName("pushButton_next_job")
        self.horizontalLayout_3.addWidget(self.pushButton_next_job)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem)
        self.box_automatic = QtWidgets.QCheckBox(self.control_area)
        self.box_automatic.setObjectName("box_automatic")
        self.horizontalLayout_3.addWidget(self.box_automatic)
        self.tool_line.addWidget(self.control_area)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.tool_line.addItem(spacerItem1)
        self.label_current_progress = QtWidgets.QLabel(self.centralwidget)
        self.label_current_progress.setMinimumSize(QtCore.QSize(130, 0))
        self.label_current_progress.setText("")
        self.label_current_progress.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_current_progress.setObjectName("label_current_progress")
        self.tool_line.addWidget(self.label_current_progress)
        self.progressBar_current = QtWidgets.QProgressBar(self.centralwidget)
        self.progressBar_current.setProperty("value", 0)
        self.progressBar_current.setObjectName("progressBar_current")
        self.tool_line.addWidget(self.progressBar_current)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.tool_line.addItem(spacerItem2)
        self.label_batch_progress = QtWidgets.QLabel(self.centralwidget)
        self.label_batch_progress.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_batch_progress.setObjectName("label_batch_progress")
        self.tool_line.addWidget(self.label_batch_progress)
        self.progressBar_batch = QtWidgets.QProgressBar(self.centralwidget)
        self.progressBar_batch.setProperty("value", 0)
        self.progressBar_batch.setObjectName("progressBar_batch")
        self.tool_line.addWidget(self.progressBar_batch)
        spacerItem3 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.tool_line.addItem(spacerItem3)
        self.pushButton_quit = QtWidgets.QPushButton(self.centralwidget)
        self.pushButton_quit.setObjectName("pushButton_quit")
        self.tool_line.addWidget(self.pushButton_quit)
        self.tool_line.setStretch(1, 3)
        self.tool_line.setStretch(2, 5)
        self.tool_line.setStretch(3, 12)
        self.tool_line.setStretch(4, 3)
        self.tool_line.setStretch(5, 3)
        self.tool_line.setStretch(6, 12)
        self.tool_line.setStretch(7, 1)
        self.verticalLayout_2.addLayout(self.tool_line)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 1098, 21))
        self.menubar.setObjectName("menubar")
        self.menuFile = QtWidgets.QMenu(self.menubar)
        self.menuFile.setObjectName("menuFile")
        self.menuEdit = QtWidgets.QMenu(self.menubar)
        self.menuEdit.setObjectName("menuEdit")
        self.menuCalibrate = QtWidgets.QMenu(self.menubar)
        self.menuCalibrate.setObjectName("menuCalibrate")
        self.menuAbout_PSS = QtWidgets.QMenu(self.menubar)
        self.menuAbout_PSS.setObjectName("menuAbout_PSS")
        MainWindow.setMenuBar(self.menubar)
        self.statusBar = QtWidgets.QStatusBar(MainWindow)
        self.statusBar.setObjectName("statusBar")
        MainWindow.setStatusBar(self.statusBar)
        self.actionLoad_video_directory = QtWidgets.QAction(MainWindow)
        self.actionLoad_video_directory.setObjectName("actionLoad_video_directory")
        self.actionLoad_config = QtWidgets.QAction(MainWindow)
        self.actionLoad_config.setObjectName("actionLoad_config")
        self.actionSave_config = QtWidgets.QAction(MainWindow)
        self.actionSave_config.setObjectName("actionSave_config")
        self.actionLoad_postproc_config = QtWidgets.QAction(MainWindow)
        self.actionLoad_postproc_config.setObjectName("actionLoad_postproc_config")
        self.actionSave_postproc_config = QtWidgets.QAction(MainWindow)
        self.actionSave_postproc_config.setObjectName("actionSave_postproc_config")
        self.actionSave = QtWidgets.QAction(MainWindow)
        self.actionSave.setObjectName("actionSave")
        self.actionSave_as = QtWidgets.QAction(MainWindow)
        self.actionSave_as.setObjectName("actionSave_as")
        self.actionQuit = QtWidgets.QAction(MainWindow)
        self.actionQuit.setObjectName("actionQuit")
        self.actionEdit_configuration = QtWidgets.QAction(MainWindow)
        self.actionEdit_configuration.setObjectName("actionEdit_configuration")
        self.actionEdit_postproc_config = QtWidgets.QAction(MainWindow)
        self.actionEdit_postproc_config.setObjectName("actionEdit_postproc_config")
        self.actionDe_activate_master_frames = QtWidgets.QAction(MainWindow)
        self.actionDe_activate_master_frames.setObjectName("actionDe_activate_master_frames")
        self.actionLoad_master_dark_frame = QtWidgets.QAction(MainWindow)
        self.actionLoad_master_dark_frame.setObjectName("actionLoad_master_dark_frame")
        self.actionLoad_master_flat_frame = QtWidgets.QAction(MainWindow)
        self.actionLoad_master_flat_frame.setObjectName("actionLoad_master_flat_frame")
        self.actionCreate_new_master_dark_frame = QtWidgets.QAction(MainWindow)
        self.actionCreate_new_master_dark_frame.setObjectName("actionCreate_new_master_dark_frame")
        self.actionCreate_new_master_flat_frame = QtWidgets.QAction(MainWindow)
        self.actionCreate_new_master_flat_frame.setObjectName("actionCreate_new_master_flat_frame")
        self.actionAbout = QtWidgets.QAction(MainWindow)
        self.actionAbout.setObjectName("actionAbout")
        self.actionShow_Quickstart = QtWidgets.QAction(MainWindow)
        self.actionShow_Quickstart.setCheckable(True)
        self.actionShow_Quickstart.setChecked(True)
        self.actionShow_Quickstart.setObjectName("actionShow_Quickstart")
        self.menuFile.addAction(self.actionLoad_video_directory)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionSave)
        self.menuFile.addAction(self.actionSave_as)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionLoad_config)
        self.menuFile.addAction(self.actionSave_config)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionQuit)
        self.menuEdit.addAction(self.actionEdit_configuration)
        self.menuCalibrate.addAction(self.actionDe_activate_master_frames)
        self.menuCalibrate.addSeparator()
        self.menuCalibrate.addAction(self.actionLoad_master_dark_frame)
        self.menuCalibrate.addAction(self.actionLoad_master_flat_frame)
        self.menuCalibrate.addSeparator()
        self.menuCalibrate.addAction(self.actionCreate_new_master_dark_frame)
        self.menuCalibrate.addAction(self.actionCreate_new_master_flat_frame)
        self.menuAbout_PSS.addAction(self.actionAbout)
        self.menuAbout_PSS.addAction(self.actionShow_Quickstart)
        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuEdit.menuAction())
        self.menubar.addAction(self.menuCalibrate.menuAction())
        self.menubar.addAction(self.menuAbout_PSS.menuAction())

        self.retranslateUi(MainWindow)
        self.comboBox_back.setCurrentIndex(-1)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "Planetary System Stacker"))
        self.comboBox_back.setToolTip(_translate("MainWindow", "Repeat starting from an earlier processing phase, or go back to the previous job."))
        self.pushButton_pause.setToolTip(_translate("MainWindow", "Suspend job execution after the current step"))
        self.pushButton_pause.setText(_translate("MainWindow", "Pause"))
        self.pushButton_start.setToolTip(_translate("MainWindow", "Start or continue job execution."))
        self.pushButton_start.setText(_translate("MainWindow", "Start / Cont."))
        self.pushButton_next_job.setToolTip(_translate("MainWindow", "Skip the remaining phases of the current job and start with the next one."))
        self.pushButton_next_job.setText(_translate("MainWindow", "Next Job"))
        self.box_automatic.setToolTip(_translate("MainWindow", "Continue in automatic (batch) mode."))
        self.box_automatic.setText(_translate("MainWindow", "Automatic"))
        self.label_batch_progress.setText(_translate("MainWindow", "Batch progress"))
        self.pushButton_quit.setToolTip(_translate("MainWindow", "Exit the program. The current configuration will be saved for the next run."))
        self.pushButton_quit.setText(_translate("MainWindow", "Quit"))
        self.menuFile.setTitle(_translate("MainWindow", "File"))
        self.menuEdit.setTitle(_translate("MainWindow", "Edit"))
        self.menuCalibrate.setTitle(_translate("MainWindow", "Calibrate"))
        self.menuAbout_PSS.setTitle(_translate("MainWindow", "Info"))
        self.actionLoad_video_directory.setText(_translate("MainWindow", "Open"))
        self.actionLoad_config.setText(_translate("MainWindow", "Load configuration"))
        self.actionLoad_config.setToolTip(_translate("MainWindow", "Load configuration from \".pss\" file"))
        self.actionSave_config.setText(_translate("MainWindow", "Save configuration"))
        self.actionSave_config.setToolTip(_translate("MainWindow", "Save configuration to \".pss\" file"))
        self.actionLoad_postproc_config.setText(_translate("MainWindow", "Load postproc config"))
        self.actionSave_postproc_config.setText(_translate("MainWindow", "Save postproc config"))
        self.actionSave.setText(_translate("MainWindow", "Save"))
        self.actionSave_as.setText(_translate("MainWindow", "Save as"))
        self.actionQuit.setText(_translate("MainWindow", "Quit"))
        self.actionEdit_configuration.setText(_translate("MainWindow", "Edit configuration"))
        self.actionEdit_postproc_config.setText(_translate("MainWindow", "Edit postproc config"))
        self.actionDe_activate_master_frames.setText(_translate("MainWindow", "De-activate master frames"))
        self.actionLoad_master_dark_frame.setText(_translate("MainWindow", "Load master dark frame"))
        self.actionLoad_master_flat_frame.setText(_translate("MainWindow", "Load master flat frame"))
        self.actionCreate_new_master_dark_frame.setText(_translate("MainWindow", "Create new master dark frame"))
        self.actionCreate_new_master_flat_frame.setText(_translate("MainWindow", "Create new master flat frame"))
        self.actionAbout.setText(_translate("MainWindow", "About PSS"))
        self.actionShow_Quickstart.setText(_translate("MainWindow", "Show quickstart"))



# -*- coding: utf-8; -*-
"""
Copyright (c) 2018 Rolf Hempel, rolf6419@gmx.de

This file is part of the PlanetarySystemStacker tool (PSS).
https://github.com/Rolf-Hempel/PlanetarySystemStacker

PSS is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PSS.  If not, see <http://www.gnu.org/licenses/>.

"""

from copy import copy, deepcopy
from math import sqrt
from pathlib import Path
from sys import argv, stdout, stderr
from time import sleep

import psutil
from PyQt5 import QtWidgets, QtCore
from PyQt5.QtWidgets import QProxyStyle, QStyle
from cv2 import imread, cvtColor, COLOR_BGR2RGB, GaussianBlur, bilateralFilter, BORDER_DEFAULT, \
    COLOR_BGR2HSV, COLOR_HSV2BGR
from numpy import uint8, uint16, float32

from configuration import Configuration, PostprocLayer
from exceptions import InternalError
from frame_viewer import FrameViewer
from frames import Frames
from miscellaneous import Miscellaneous
from postproc_editor_gui import Ui_postproc_editor
from sharpening_layer_widget import Ui_sharpening_layer_widget
from version_manager_widget import Ui_version_manager_widget


class SharpeningLayerWidget(QtWidgets.QWidget, Ui_sharpening_layer_widget):
    """
    GUI widget to manipulate the parameters of a sharpening layer. An arbitrary number of sharpening
    layers can be defined (up to a maximum set as configuration parameter).
    """
    signal_toggle_luminance_other_layers = QtCore.pyqtSignal()

    def __init__(self, layer_index, remove_layer_callback, parent=None):
        """
        Initialize the widget.

        :param layer_index: Index of the sharpening layer (starting with 1).
        :param remove_layer_callback: Callback function to remove a layer from the current version.
        :param parent: Parent object.
        """

        QtWidgets.QWidget.__init__(self, parent)
        self.setupUi(self)

        self.layer_index = layer_index
        self.title = "Layer " + str(layer_index + 1)
        self.remove_layer_callback = remove_layer_callback

        self.horizontalSlider_radius.valueChanged.connect(self.horizontalSlider_radius_changed)
        self.lineEdit_radius.textChanged.connect(self.lineEdit_radius_changed)
        self.horizontalSlider_amount.valueChanged.connect(self.horizontalSlider_amount_changed)
        self.lineEdit_amount.textChanged.connect(self.lineEdit_amount_changed)
        self.horizontalSlider_bi_fraction.valueChanged.connect(self.horizontalSlider_bi_fraction_changed)
        self.lineEdit_bi_fraction.textChanged.connect(self.lineEdit_bi_fraction_changed)
        self.horizontalSlider_bi_range.valueChanged.connect(
            self.horizontalSlider_bi_range_changed)
        self.lineEdit_bi_range.textChanged.connect(self.lineEdit_bi_range_changed)
        self.horizontalSlider_denoise.valueChanged.connect(
            self.horizontalSlider_denoise_changed)
        self.lineEdit_denoise.textChanged.connect(self.lineEdit_denoise_changed)
        self.checkBox_luminance.stateChanged.connect(self.checkBox_luminance_toggled)
        self.pushButton_remove.clicked.connect(self.remove_layer)

    def set_values(self, layer):
        """
        Set the values of the GUI widget according to the current data model

        :param layer: Layer instance holding the sharpening parameters
        :return: -
        """

        self.layer = layer

        self.groupBox_layer.setTitle(self.title)
        self.horizontalSlider_radius.setValue(self.radius_to_integer(self.layer.radius))
        self.lineEdit_radius.setText(str(self.layer.radius))
        self.horizontalSlider_amount.setValue(self.amount_to_integer(self.layer.amount))
        self.lineEdit_amount.setText("{0:.2f}".format(round(self.layer.amount, 2)))
        self.horizontalSlider_bi_fraction.setValue(
            self.bi_fraction_to_integer(self.layer.bi_fraction))
        self.lineEdit_bi_fraction.setText("{0:.2f}".format(round(self.layer.bi_fraction, 2)))
        self.horizontalSlider_bi_range.setValue(self.bi_range_to_integer(self.layer.bi_range))
        self.lineEdit_bi_range.setText(str(self.layer.bi_range))
        self.horizontalSlider_denoise.setValue(
            self.denoise_to_integer(self.layer.denoise))
        self.lineEdit_denoise.setText("{0:.2f}".format(round(self.layer.denoise, 2)))

        # Temporarily block signals for the luminance checkbox. Otherwise the variable would be
        # switched immediately.
        self.checkBox_luminance.blockSignals(True)
        self.checkBox_luminance.setChecked(self.layer.luminance_only)
        self.checkBox_luminance.blockSignals(False)

    # The following four methods implement the translation between data model values and the
    # (integer) values of the horizontal GUI sliders. In the case of the sharpening amount this
    # translation is non-linear in order to add resolution at small values.
    @staticmethod
    def radius_to_integer(radius):
        return max(min(round(radius * 10.), 99), 1)

    @staticmethod
    def integer_to_radius(int):
        return int / 10.

    @staticmethod
    def amount_to_integer(amount):
        # Below the slider value 20 (amount = 1.) the behaviour is linear, above quadratic.
        if amount > 1.:
            a = 187. / 6400.
            b = 0.15 - 40. * a
            c = 400. * a - 2.
            return (round(-b / (2. * a) + sqrt(b**2 / a**2 / 4. - (c - amount) / a)))
        else:
            return round((amount + 2.) / 0.15)

    @staticmethod
    def integer_to_amount(integer):
        if integer <= 20:
            return -2. + 0.15 * integer
        else:
            a = 187. / 6400.
            b = 0.15 - 40. * a
            c = 400. * a - 2.
            return round(a * integer**2 + b * integer + c, 2)

    @staticmethod
    def bi_fraction_to_integer(bi_fraction):
        return round(bi_fraction * 100)

    @staticmethod
    def integer_to_bi_fraction(integer):
        return integer / 100.

    @staticmethod
    def bi_range_to_integer(bi_range):
        # Below the slider value 50 (bi_range = 13.) the behaviour is linear, above quadratic.
        if bi_range > 13.:
            c = 0.009022222222222221
            g = 35.59113300492611
            h = -1233.2712514256596
            return round(g + sqrt(h + bi_range / c))
        else:
            b = 0.26
            return round(bi_range / b)

    @staticmethod
    def integer_to_bi_range(integer):
        if integer <= 50:
            b = 0.26
            return round(b * integer, 1)
        else:
            c = 0.009022222222222221
            e = -0.6422222222222221
            f = 22.555555555555554
            return round(c * integer ** 2 + e * integer + f, 1)

    @staticmethod
    def denoise_to_integer(denoise):
        return round(denoise * 100)

    @staticmethod
    def integer_to_denoise(integer):
        return integer / 100.

    def horizontalSlider_radius_changed(self):
        """
        The "radius" slider has been moved. Update the corresponding lineEdit widget. Block its
        signals temporarily to avoid cross-talk.
        :return: -
        """

        self.layer.radius = self.integer_to_radius(self.horizontalSlider_radius.value())
        self.lineEdit_radius.blockSignals(True)
        self.lineEdit_radius.setText(str(self.layer.radius))
        self.lineEdit_radius.blockSignals(False)

    def lineEdit_radius_changed(self):
        """
        The text of the lineEdit widget has been changed. Update the horizontal slider. Include a
        try-except block to avoid illegal user input to crash the program.

        :return: -
        """

        try:
            self.layer.radius = max(0.1, min(float(self.lineEdit_radius.text()), 9.9))
            self.horizontalSlider_radius.blockSignals(True)
            self.horizontalSlider_radius.setValue(self.radius_to_integer(self.layer.radius))
            self.horizontalSlider_radius.blockSignals(False)
        except:
            pass

    def horizontalSlider_amount_changed(self):
        """
        The same as above for the "amount" parameter.

        :return: -
        """

        self.layer.amount = self.integer_to_amount(self.horizontalSlider_amount.value())
        self.lineEdit_amount.blockSignals(True)
        self.lineEdit_amount.setText("{0:.2f}".format(round(self.layer.amount, 2)))
        self.lineEdit_amount.blockSignals(False)

    def lineEdit_amount_changed(self):
        """
        The same as above for the "amount" parameter.

        :return: -
        """

        try:
            self.layer.amount = float(self.lineEdit_amount.text())
            self.horizontalSlider_amount.blockSignals(True)
            self.horizontalSlider_amount.setValue(self.amount_to_integer(self.layer.amount))
            self.horizontalSlider_amount.blockSignals(False)
        except:
            pass

    def horizontalSlider_bi_fraction_changed(self):
        """
        The same as above for the "bi_fraction" parameter.

        :return: -
        """

        self.layer.bi_fraction = self.integer_to_bi_fraction(self.horizontalSlider_bi_fraction.value())
        self.lineEdit_bi_fraction.blockSignals(True)
        self.lineEdit_bi_fraction.setText("{0:.2f}".format(round(self.layer.bi_fraction, 2)))
        self.lineEdit_bi_fraction.blockSignals(False)

    def lineEdit_bi_fraction_changed(self):
        """
        The same as above for the "bi_fraction" parameter.

        :return: -
        """

        try:
            self.layer.bi_fraction = float(self.lineEdit_bi_fraction.text())
            self.horizontalSlider_bi_fraction.blockSignals(True)
            self.horizontalSlider_bi_fraction.setValue(self.bi_fraction_to_integer(self.layer.bi_fraction))
            self.horizontalSlider_bi_fraction.blockSignals(False)
        except:
            pass

    def horizontalSlider_bi_range_changed(self):
        """
        The same as above for the "bi_range" parameter.

        :return: -
        """

        self.layer.bi_range = self.integer_to_bi_range(self.horizontalSlider_bi_range.value())
        self.lineEdit_bi_range.blockSignals(True)
        self.lineEdit_bi_range.setText(str(self.layer.bi_range))
        self.lineEdit_bi_range.blockSignals(False)

    def lineEdit_bi_range_changed(self):
        """
        The same as above for the "bi_range" parameter.

        :return: -
        """

        try:
            self.layer.bi_range = max(0, min(int(round(float(self.lineEdit_bi_range.text()))), 255))
            self.horizontalSlider_bi_range.blockSignals(True)
            self.horizontalSlider_bi_range.setValue(self.bi_range_to_integer(self.layer.bi_range))
            self.horizontalSlider_bi_range.blockSignals(False)
        except:
            pass

    def horizontalSlider_denoise_changed(self):
        """
        The same as above for the "denoise" parameter.

        :return: -
        """

        self.layer.denoise = self.integer_to_denoise(self.horizontalSlider_denoise.value())
        self.lineEdit_denoise.blockSignals(True)
        self.lineEdit_denoise.setText("{0:.2f}".format(round(self.layer.denoise, 2)))
        self.lineEdit_denoise.blockSignals(False)

    def lineEdit_denoise_changed(self):
        """
        The same as above for the "denoise" parameter.

        :return: -
        """

        try:
            self.layer.denoise = float(self.lineEdit_denoise.text())
            self.horizontalSlider_denoise.blockSignals(True)
            self.horizontalSlider_denoise.setValue(self.denoise_to_integer(self.layer.denoise))
            self.horizontalSlider_denoise.blockSignals(False)
        except:
            pass


    def checkBox_luminance_toggled(self):
        """
        The checkbox which tells if the sharpening should affect the luminance channel only has
        changed its state.

        :return: -
        """

        self.layer.luminance_only = not self.layer.luminance_only

        # Signal the other layers to change the flag as well.
        self.signal_toggle_luminance_other_layers.emit()

    @QtCore.pyqtSlot()
    def checkBox_luminance_toggled_no_signal(self):
        """
        The checkbox which tells if the sharpening should affect the luminance channel only has
        changed its state. This version of the method is invoked by another layer which has changed
        the toggle state. In this case, not only toggle the state variable, but also toggle the
        checkbox widget (because this was not done by the user). To avoid an avalanche, do not emit
        a signal.

        :return: -
        """

        # Block the signals of this widget to avoid a feedback loop.
        self.checkBox_luminance.blockSignals(True)
        self.checkBox_luminance.setChecked(not self.checkBox_luminance.isChecked())
        self.checkBox_luminance.blockSignals(False)
        self.layer.luminance_only = not self.layer.luminance_only

    def remove_layer(self):
        """
        This method is executed when the user presses the "remove" button. The actual action is
        performed by the higher-level "PostprocEditorWidget" object.
        :return: -
        """

        self.remove_layer_callback(self.layer_index)


class CustomStyle(QProxyStyle):
    """
    This class is used to prevent the version spinbox from jumping two steps at once. The solution
    was found on Stackoverflow
    (https://stackoverflow.com/questions/40746350/why-qspinbox-jumps-twice-the-step-value).

    """

    def styleHint(self, hint, option=None, widget=None, returnData=None):
        if hint == QStyle.SH_SpinBox_KeyPressAutoRepeatRate:
            return 10**6
        elif hint == QStyle.SH_SpinBox_ClickAutoRepeatRate:
            return 10**6
        elif hint == QStyle.SH_SpinBox_ClickAutoRepeatThreshold:
            # You can use only this condition to avoid the auto-repeat,
            # but better safe than sorry ;-)
            return 10**6
        else:
            return super().styleHint(hint, option, widget, returnData)


class VersionManagerWidget(QtWidgets.QWidget, Ui_version_manager_widget):
    """
    This GUI widget manages the different postprocessing versions and controls which image is
    displayed in the image viewer. It also starts and controls the blink comparator.
    """

    # The set_photo_signal tells the image viewer which image it should show.
    set_photo_signal = QtCore.pyqtSignal(int)
    # The blink comparator emits the variant_shown_signal to highlight / de-highlight the GUI image
    # selector which corresponds to the image currently displayed.
    variant_shown_signal = QtCore.pyqtSignal(bool, bool)
    # This signal tells the PostprocEditorWidget to show the selected version in the frame viewer.
    select_version_signal = QtCore.pyqtSignal(int)

    def __init__(self, configuration, parent=None):
        """

        :param configuration: Configuration object with parameters.
        :param parent: parent object.
        """

        QtWidgets.QWidget.__init__(self, parent)
        self.setupUi(self)
        self.configuration = configuration
        self.pss_version = configuration.global_parameters_version
        self.postproc_data_object = configuration.postproc_data_object
        self.postproc_blinking_period = configuration.postproc_blinking_period

        self.spinBox_version.valueChanged.connect(self.select_version)
        self.spinBox_compare.valueChanged.connect(self.select_version_compared)
        self.pushButton_new.clicked.connect(self.new_version)
        self.pushButton_delete.clicked.connect(self.remove_version)
        self.checkBox_blink_compare.stateChanged.connect(self.blinking_toggled)
        self.pushButton_save.clicked.connect(self.save_version)
        self.pushButton_save_as.clicked.connect(self.save_version_as)
        self.variant_shown_signal.connect(self.highlight_variant)

        # Set the ranges for version selection spinboxes. Initially, there are only two versions.
        self.spinBox_version.setMaximum(configuration.postproc_data_object.number_versions)
        self.spinBox_version.setMinimum(0)
        self.spinBox_compare.setMaximum(configuration.postproc_data_object.number_versions)
        self.spinBox_compare.setMinimum(0)
        self.spinBox_version.setStyle(CustomStyle())
        self.spinBox_compare.setStyle(CustomStyle())

        # Set the spinbox to the newly created version.
        self.spinBox_version.setValue(self.postproc_data_object.version_selected)

    def select_version(self):
        """
        This method is called when the user changes the setting of the "version selected" spinbox.

        :return: -
        """
        self.postproc_data_object.version_selected = self.spinBox_version.value()
        self.select_version_signal.emit(self.postproc_data_object.version_selected)

    def select_version_compared(self):
        """
        This method is called when the user changes the setting of the "version compared" spinbox.
        The "selected" and "compared" versions are displayed alternately by the blink comparator.
        :return: -
        """

        self.postproc_data_object.version_compared = self.spinBox_compare.value()

    def new_version(self):
        """
        Create a new postprocessing version.
        :return: -
        """

        # Create a Version object by copying the parameters from the selected version.
        new_version = self.postproc_data_object.new_postproc_version_from_existing()

        # If the new version was created from the original image (version 0), add an initial layer.
        if self.postproc_data_object.version_selected == 1:
            new_version.add_postproc_layer(PostprocLayer("Multilevel unsharp masking", 1., 1., 0.,
                                    self.configuration.postproc_bi_range_standard, 0., False))

        # Set the image viewer to the new version, and increase the range of spinboxes to include
        # the new version.
        self.set_photo_signal.emit(self.postproc_data_object.version_selected)
        self.spinBox_version.setMaximum(self.postproc_data_object.number_versions)
        self.spinBox_version.setValue(self.postproc_data_object.number_versions)
        self.spinBox_compare.setMaximum(self.postproc_data_object.number_versions)

    def remove_version(self):
        """
        Remove a postprocessing version.
        :return: -
        """

        # Remove the version from the central data object, and instruct the image viewer to
        # show the previous version.
        self.postproc_data_object.remove_postproc_version(self.postproc_data_object.version_selected)
        self.set_photo_signal.emit(self.postproc_data_object.version_selected)

        # Adjust the spinbox bounds, and set the current version parameters to the new selection.
        self.spinBox_version.setMaximum(self.postproc_data_object.number_versions)

        # Set the spinbox to the version before the deleted one.
        self.spinBox_version.setValue(self.postproc_data_object.version_selected)

        self.spinBox_compare.setMaximum(self.postproc_data_object.number_versions)
        self.spinBox_compare.setValue(
            min(self.spinBox_compare.value(), self.postproc_data_object.number_versions))
        self.select_version_signal.emit(self.postproc_data_object.version_selected)

    def blinking_toggled(self):
        """
        Switch the blink comparator on or off.

        :return: -
        """

        # Toggle the status variable.
        self.postproc_data_object.blinking = not self.postproc_data_object.blinking

        # The blink comparator is switched on.
        if self.postproc_data_object.blinking:
            # Create the blink comparator thread and start it.
            self.blink_comparator = BlinkComparator(self.postproc_data_object,
                                                    self.postproc_blinking_period,
                                                    self.set_photo_signal,
                                                    self.variant_shown_signal)
            self.blink_comparator.setTerminationEnabled(True)

        # The blink comparator is switched off. Set the image viewer to the currently selected
        # image and terminate the separate thread.
        else:
            self.set_photo_signal.emit(self.postproc_data_object.version_selected)
            self.blink_comparator.stop()

    def highlight_variant(self, selected, compare):
        """
        Whenever the blink comparator changes the display, the spinbox corressponding to the
        currently displayed image changes its font color to red. The other spinbox is blanked out
        (by setting font color to white). When the blink comparator is not active, both spinboxes
        show their numbers in bladk font.

        :param selected: True, if the image corresponding to the "selected" spinbox is displayed.
                         False, otherwise.
        :param compare: True, if the image corresponding to the "compared" spinbox is displayed.
                        False, otherwise.
        :return: -
        """

        if selected and compare:
            self.spinBox_version.setStyleSheet('color: red')
            self.spinBox_compare.setStyleSheet('color: red')
        elif not selected and not compare:
            self.spinBox_version.setStyleSheet('color: black')
            self.spinBox_compare.setStyleSheet('color: black')
        elif selected:
            self.spinBox_version.setStyleSheet('color: red')
            self.spinBox_compare.setStyleSheet('color: white')
        elif compare:
            self.spinBox_version.setStyleSheet('color: white')
            self.spinBox_compare.setStyleSheet('color: red')

    def save_version(self):
        """
        save the result as 16bit png, tiff or fits file at the standard location.

        :return: -
        """

        # In case "correction mode" is on, process the current version image in full 16bit
        # resolution.
        self.postproc_data_object.finalize_postproc_version()
        Frames.save_image(self.postproc_data_object.file_name_processed,
                          self.postproc_data_object.versions[
                              self.postproc_data_object.version_selected].image,
                          color=self.postproc_data_object.color, avoid_overwriting=False,
                          header=self.pss_version)

    def save_version_as(self):
        """
        save the result as 16bit png, tiff or fits file at a location selected by the user.

        :return: -
        """

        options = QtWidgets.QFileDialog.Options()
        filename, extension = QtWidgets.QFileDialog.getSaveFileName(self,
                            "Save result as 16bit png, tiff or fits image",
                            self.postproc_data_object.file_name_processed,
                            "Image Files (*.png *.tiff *.fits)", options=options)

        if filename and extension:
            self.postproc_data_object.finalize_postproc_version()
            Frames.save_image(filename,
                              self.postproc_data_object.versions[
                                  self.postproc_data_object.version_selected].image,
                              color=self.postproc_data_object.color, avoid_overwriting=False,
                              header=self.pss_version)


class BlinkComparator(QtCore.QThread):
    """
    This class implements the blink comparator. It shows alternately two image versions.
    """

    def __init__(self, postproc_data_object, postproc_blinking_period, set_photo_signal,
                 variant_shown_signal, parent=None):
        """
        :param postproc_data_object: The central postprocessing data object.
        :param postproc_blinking_period: Time between image switching.
        :param set_photo_signal: Signal which tells the image viewer which image it should show.
        :param variant_shown_signal: Signal which tells the GUI widget which spinbox should be
                                     highlighted.
        :param parent: Parent object
        """

        QtCore.QThread.__init__(self, parent)
        self.postproc_data_object = postproc_data_object
        self.postproc_blinking_period = postproc_blinking_period
        self.set_photo_signal = set_photo_signal
        self.variant_shown_signal = variant_shown_signal

        # Initially, no version is shown by the blink comparator.
        self.variant_shown_signal.emit(False, False)

        # Start the comparator thread.
        self.start()

    def run(self):
        """
        Blink comparator main loop. As long as the "blinking" status variable is True, show the
        two image versions alternately.

        :return: -
        """

        # Begin with showing the "selected" version.
        show_selected_version = True

        # Begin main loop.
        while self.postproc_data_object.blinking:
            # Show the "selected" version.
            if show_selected_version:
                # In case the blink comparator is invoked during manual RGB shift corrections,
                # produce a final image of the selected version before blinking.
                self.postproc_data_object.finalize_postproc_version()
                self.set_photo_signal.emit(self.postproc_data_object.version_selected)
                self.variant_shown_signal.emit(True, False)
            # Show the "compared" version.
            else:
                self.postproc_data_object.finalize_postproc_version(
                    version_index=self.postproc_data_object.version_compared)
                self.set_photo_signal.emit(self.postproc_data_object.version_compared)
                self.variant_shown_signal.emit(False, True)

            # Toggle back and forth between first and second image version.
            show_selected_version = not show_selected_version

            # Sleep time inserted to limit CPU consumption by idle looping.
            sleep(self.postproc_blinking_period)

    def stop(self):
        """
        Terminate the blink comparator thread.

        :return: -
        """

        self.variant_shown_signal.emit(False, False)
        self.terminate()


class ImageProcessor(QtCore.QThread):
    """
    This class implements the asynchronous computation of new image versions.
    """

    # The set_photo_signal tells the image viewer which image it should show.
    set_photo_signal = QtCore.pyqtSignal(int)
    # The set_shift_display_signal tells the PostprocEditorWidget to update the display of RGB
    # shift values.
    set_shift_display_signal = QtCore.pyqtSignal()
    # The set_status_bar_signal triggers the display of a (colored) message in the main GUI's
    # status bar.
    set_status_bar_signal = QtCore.pyqtSignal(str, str)
    # While computations are going on, GUI widgets are disabled to avoid race conditions.
    disable_widgets_signal = QtCore.pyqtSignal()
    # When computations have finished, reactivate GUI widgets.
    enable_widgets_signal = QtCore.pyqtSignal()

    def __init__(self, configuration, parent=None):
        """
        Whenever the parameters of the current version change, compute a new sharpened image.

        :param configuration: Data object with configuration parameters
        """

        QtCore.QThread.__init__(self, parent)
        self.configuration = configuration
        self.postproc_data_object = self.configuration.postproc_data_object
        self.postproc_idle_loop_time = self.configuration.postproc_idle_loop_time

        # Extract the file name from its path.
        self.file_name = Path(self.postproc_data_object.file_name_original).name

        self.start()

    def run(self):

        # Before executing the main loop, initialize images for all versions kept in the version
        # manager. In particular if versions have the "Auto RGB alignment" checkbox set, this may
        # take a while.

        # Change the main GUI's status bar to show that a computation is going on, and disable GUI
        # widgets to avoid race conditions.
        self.set_status_bar_signal.emit("Postprocessing " + self.file_name +
                                        ", busy computing initial images for all versions; "
                                        "This will take a while",
                                        "black")
        self.disable_widgets_signal.emit()

        # Do the computations in float32 to avoid clipping effects. If the input image is color,
        # also create a version for "luminance only" computations.
        self.image_original = self.postproc_data_object.image_original.astype(float32)
        self.color = len(self.image_original.shape) == 3
        if self.color:
            self.image_original_hsv = cvtColor(self.image_original, COLOR_BGR2HSV)

        # Initialize list of auto-rgb-aligned input images for all resolution levels.
        self.auto_rgb_aligned_images_original = [None, None, None]
        if self.color:
            self.auto_rgb_aligned_images_original_hsv = [None, None, None]
        self.auto_rgb_shifts_red = [None, None, None]
        self.auto_rgb_shifts_blue = [None, None, None]

        # Initialize images for all versions using the current layer data.
        self.last_version_layers = []

        for version in self.postproc_data_object.versions:
            # For every version, keep a copy of the current layer parameters for later checks for
            # changes. Also, remember if for this version RGB alignment was active.
            self.last_version_layers.append(deepcopy(version.layers))
            version.last_rgb_automatic = version.rgb_automatic

            # If automatic RGB alignment is selected for this version, compute the shifted
            # original image for this version's resolution if not yet available.
            if version.rgb_automatic:
                if self.auto_rgb_aligned_images_original[version.rgb_resolution_index] is None:
                    try:
                        self.auto_rgb_aligned_images_original[version.rgb_resolution_index], \
                        self.auto_rgb_shifts_red[version.rgb_resolution_index], \
                        self.auto_rgb_shifts_blue[
                            version.rgb_resolution_index] = Miscellaneous.auto_rgb_align(
                            self.image_original, self.configuration.postproc_max_shift,
                            interpolation_factor=[1, 2, 4][version.rgb_resolution_index],
                            blur_strength=version.rgb_gauss_width)
                        if self.color:
                            self.auto_rgb_aligned_images_original_hsv[
                                version.rgb_resolution_index] = cvtColor(
                                self.auto_rgb_aligned_images_original[version.rgb_resolution_index],
                                COLOR_BGR2HSV)
                    except:
                        self.auto_rgb_aligned_images_original[
                            version.rgb_resolution_index] = self.image_original
                        self.auto_rgb_shifts_red[version.rgb_resolution_index] = (0., 0.)
                        self.auto_rgb_shifts_blue[version.rgb_resolution_index] = (0., 0.)
                        if self.color:
                            self.auto_rgb_aligned_images_original_hsv[
                                version.rgb_resolution_index] = self.image_original_hsv

                version.image = Miscellaneous.post_process(
                    self.auto_rgb_aligned_images_original[version.rgb_resolution_index],
                    version.layers)
            # If shifts were set manually, apply them before sharpening the input image.
            elif version.shift_red != (0., 0.) or version.shift_blue != (0., 0.):
                # Shift the image with the resolution given by the selected interpolation factor.
                interpolation_factor = [1, 2, 4][version.rgb_resolution_index]
                shifted_image = Miscellaneous.shift_colors(self.image_original,
                                                          version.shift_red, version.shift_blue,
                                                          interpolate_input=interpolation_factor,
                                                          reduce_output=interpolation_factor)
                version.image = Miscellaneous.post_process(shifted_image, version.layers)
            # No RGB corrections were specified.
            else:
                version.image = Miscellaneous.post_process(self.image_original, version.layers)

        # Initialize lists for intermediate results (to speed up image updates). The efficient
        # reuse of intermediate results is only possible if there is enough RAM.
        self.reset_intermediate_images()
        self.ram_sufficient = True

        # Reset the status bar to its idle state, and enable GUI widgets.
        self.set_status_bar_signal.emit("Postprocessing " + self.file_name,
                                        "black")
        self.enable_widgets_signal.emit()

        # Remember the last version (and the corresponding layer parameters) shown in the image
        # viewer. As soon as this index changes, a new image is displayed.
        self.last_version_selected = 0

        # Initialize the data objects holding the currently active and the previous version.
        self.version_selected = None
        self.layers_selected = None

        # Initialize Exception state.
        exception_status = False

        # Enter the main loop and wait for parameter changes in the version selected.
        while True:

            # The whole computation is enclosed in a try/except block. If something goes wrong, e.g.
            # because during the execution of a loop pass the postproc_data_object was changed by
            # the GUI thread, the ImageProcessor tries again in the next pass.
            try:
                # To avoid chasing a moving target, copy the parameters of the currently active version
                self.version_selected = self.postproc_data_object.version_selected
                postproc_version = deepcopy(self.postproc_data_object.versions[self.version_selected])
                self.layers_selected = postproc_version.layers
                self.rgb_automatic = postproc_version.rgb_automatic
                self.rgb_resolution_index = postproc_version.rgb_resolution_index
                self.rgb_gauss_width = postproc_version.rgb_gauss_width
                self.shift_red = postproc_version.shift_red
                self.shift_blue = postproc_version.shift_blue
                self.correction_red = postproc_version.correction_red
                self.correction_blue = postproc_version.correction_blue
                self.correction_red_saved = postproc_version.correction_red_saved
                self.correction_blue_saved = postproc_version.correction_blue_saved

                compute_new_image = False
                shift_image = False

                # If the RGB auto-alignment checkbox was changed since the last image was computed for
                # this version, invalidate all intermediate results for this version.
                if self.rgb_automatic != self.postproc_data_object.versions[
                    self.version_selected].last_rgb_automatic:
                    self.postproc_data_object.versions[
                        self.version_selected].last_rgb_automatic = self.rgb_automatic
                    self.reset_intermediate_images()
                    compute_new_image = True

                # If automatic RGB alignment is on, check if the shifted image for the current
                # resolution has been computed already. Otherwise, compute it now.
                if self.rgb_automatic:
                    if self.auto_rgb_aligned_images_original[
                        self.rgb_resolution_index] is None:

                        # Change the main GUI's status bar to show that a computation is going on.
                        self.set_status_bar_signal.emit("Postprocessing " + self.file_name +
                                                        ", busy auto-aligning image", "black")
                        self.disable_widgets_signal.emit()
                        try:
                            self.auto_rgb_aligned_images_original[self.rgb_resolution_index], \
                            self.auto_rgb_shifts_red[self.rgb_resolution_index], \
                            self.auto_rgb_shifts_blue[
                                self.rgb_resolution_index] = Miscellaneous.auto_rgb_align(
                                self.image_original, self.configuration.postproc_max_shift,
                                interpolation_factor=[1, 2, 4][self.rgb_resolution_index],
                                blur_strength=self.rgb_gauss_width)
                            if self.color:
                                self.auto_rgb_aligned_images_original_hsv[
                                    self.rgb_resolution_index] = cvtColor(
                                    self.auto_rgb_aligned_images_original[self.rgb_resolution_index],
                                    COLOR_BGR2HSV)
                        except:
                            self.auto_rgb_aligned_images_original[
                                self.rgb_resolution_index] = self.image_original
                            self.auto_rgb_shifts_red[self.rgb_resolution_index] = (0., 0.)
                            self.auto_rgb_shifts_blue[self.rgb_resolution_index] = (0., 0.)
                            if self.color:
                                self.auto_rgb_aligned_images_original_hsv[
                                    self.rgb_resolution_index] = self.image_original_hsv
                        # Reset the status bar to its idle state.
                        self.set_status_bar_signal.emit("Postprocessing " + self.file_name, "black")
                        self.enable_widgets_signal.emit()

                        # Since the auto-shifted image is new, the postprocessing pipeline must be
                        # applied. (This does not seem to be necessary.)
                        # compute_new_image = True

                    # Set the processing pipeline input to the RGB aligned original image.
                    self.input_image = self.auto_rgb_aligned_images_original[self.rgb_resolution_index]
                    if self.color:
                        self.input_image_hsv = self.auto_rgb_aligned_images_original_hsv[
                                self.rgb_resolution_index]

                    # Set the RGB shifts for this version to the values computed automatically.
                    self.postproc_data_object.versions[self.version_selected].shift_red = \
                        self.auto_rgb_shifts_red[self.rgb_resolution_index]
                    self.postproc_data_object.versions[self.version_selected].shift_blue = \
                        self.auto_rgb_shifts_blue[self.rgb_resolution_index]

                # In correction mode the wavelets are computed only once, and then only the shift
                # corrections are applied.
                elif postproc_version.rgb_correction_mode:

                    # If the uncorrected image for this resolution index is not available, it must be
                    # computed. Try if an image for this version has been computed. If not, this version
                    # is new. In that case process the wavelets first (only applying the accumulated
                    # shifts), and do the shift corrections in the next pass.
                    if postproc_version.images_uncorrected[self.rgb_resolution_index] is None:

                        # Change the main GUI's status bar to show that correction mode is being
                        # initialized.
                        self.set_status_bar_signal.emit("Postprocessing " + self.file_name +
                                                        ", initializing manual alignment", "black")
                        self.disable_widgets_signal.emit()

                        if postproc_version.image is None or postproc_version.image.dtype == uint8:
                            self.input_image = Miscellaneous.shift_colors(self.image_original,
                                                                          postproc_version.shift_red,
                                                                          postproc_version.shift_blue)
                            if self.color:
                                if postproc_version.shift_red != (
                                0., 0.) or postproc_version.shift_blue != (0., 0.):
                                    self.input_image_hsv = cvtColor(self.input_image, COLOR_BGR2HSV)
                                else:
                                    self.input_image_hsv = self.image_original_hsv

                            self.reset_intermediate_images()
                            compute_new_image = True

                        # The postprocessing pipeline has been applied to this version before, set the
                        # input for this resolution index of the correction mode. The image is
                        # blown up by the resolution factor to show shifts in detail.
                        else:
                            # In the first pass, convert the input image to 8bit unsigned int (to speed
                            # up the image viewer). Store it as resolution version 0 (1 Pixel, i.e. no
                            # change in resolution).
                            if self.postproc_data_object.versions[
                                self.version_selected].images_uncorrected[0] is None:
                                self.postproc_data_object.versions[
                                    self.version_selected].images_uncorrected[0] = (
                                        postproc_version.image / 256.).astype(
                                    uint8)

                            # If the resolution selected is not 1 pixel, interpolate the 1 pixel image
                            # to the required resolution and store the result for later reuse.
                            if self.rgb_resolution_index > 0:
                                self.postproc_data_object.versions[
                                    self.version_selected].images_uncorrected[
                                    self.rgb_resolution_index] = \
                                    Miscellaneous.shift_colors(self.postproc_data_object.versions[
                                        self.version_selected].images_uncorrected[0], (0., 0.), (0., 0.),
                                        interpolate_input=[1, 2, 4][self.rgb_resolution_index])
                            shift_image = True

                        # Reset the status bar to its idle state.
                        self.set_status_bar_signal.emit("Postprocessing " + self.file_name, "black")
                        self.enable_widgets_signal.emit()

                    # The image without shift correction is available for the required resolution. Check
                    # if the shift correction has changed. If so, apply the correction. Otherwise, leave
                    # the image unchanged.
                    elif self.correction_red != self.correction_red_saved or self.correction_blue != \
                            self.correction_blue_saved:
                        self.postproc_data_object.versions[
                            self.version_selected].correction_red_saved = self.correction_red
                        self.postproc_data_object.versions[
                            self.version_selected].correction_blue_saved = self.correction_blue
                        shift_image = True

                # Wavelet mode, RGB auto-alignment off: Set the input image for the processing pipeline
                # to the original image, shifted by the accumulated shift vectors (not in correction
                # mode).
                else:
                    # The accumulated shift vectors have changed since the last computation. Apply the
                    # new shifts to the original image.
                    if postproc_version.shift_red != postproc_version.shift_red_saved or \
                        postproc_version.shift_blue != postproc_version.shift_blue_saved:
                        self.input_image = Miscellaneous.shift_colors(self.image_original,
                                                                      postproc_version.shift_red,
                                                                      postproc_version.shift_blue)
                        if self.color:
                            if postproc_version.shift_red != (
                                    0., 0.) or postproc_version.shift_blue != (0., 0.):
                                self.input_image_hsv = cvtColor(self.input_image, COLOR_BGR2HSV)
                            else:
                                self.input_image_hsv = self.image_original_hsv

                        # Save the new shifted original image together with the current accumulated
                        # shift vectors.
                        self.postproc_data_object.versions[self.version_selected].input_image_saved = \
                            self.input_image
                        if self.color:
                            self.postproc_data_object.versions[
                                self.version_selected].input_image_hsv_saved = self.input_image_hsv
                        self.postproc_data_object.versions[self.version_selected].shift_red_saved = \
                            postproc_version.shift_red
                        self.postproc_data_object.versions[
                            self.version_selected].shift_blue_saved = postproc_version.shift_blue

                        # Since the processing pipeline input has changed, reset intermediate results
                        # and set the computation flag.
                        self.reset_intermediate_images()
                        compute_new_image = True

                    # The accumulated shift vectors have not changed. Reuse the saved pipeline input
                    # images.
                    else:
                        self.input_image = postproc_version.input_image_saved
                        if self.color:
                            self.input_image_hsv = postproc_version.input_image_hsv_saved

                self.set_shift_display_signal.emit()

                # End of preparatory phase. Flags "compute_new_image" or "shift_image" have been set.
                # Now perform the appropriate action.

                # The image has already passed the postprocessing pipeline (as contained in
                # postproc_version.images_uncorrected). Only the correction shift must be applied.
                if shift_image:
                    interpolation_factor = [1, 2, 4][self.rgb_resolution_index]
                    self.postproc_data_object.versions[self.version_selected].image = \
                        Miscellaneous.shift_colors(
                            self.postproc_data_object.versions[
                                self.version_selected].images_uncorrected[self.rgb_resolution_index],
                            (self.correction_red[0] * interpolation_factor,
                             self.correction_red[1] * interpolation_factor),
                            (self.correction_blue[0] * interpolation_factor,
                             self.correction_blue[1] * interpolation_factor))
                    # Show the new image in the image viewer, and remember its parameters.
                    self.set_photo_signal.emit(self.version_selected)
                    self.last_version_selected = self.version_selected

                # A new image must be computed for this version. Special case self.version_selected = 0:
                # For the original image no layers are applied. But if the RGB automatic checkbox was
                # changed, the image must be set according to the new shift status. The shift was
                # applied in self.input_image above.
                elif compute_new_image and not self.version_selected:
                    self.postproc_data_object.versions[0].image = self.input_image.astype(uint16)
                    # Show the new image in the image viewer, and remember its parameters.
                    self.set_photo_signal.emit(self.version_selected)
                    self.last_version_selected = self.version_selected

                # General case for new image computation: Either it was decided above that a new
                # computation is required, or the test "new_computation_required" for changes in the
                # correction layers is performed.
                elif compute_new_image or self.new_computation_required(
                        self.version_selected != self.last_version_selected):
                    # Change the main GUI's status bar to show that a computation is going on.
                    self.set_status_bar_signal.emit("Postprocessing " + self.file_name +
                                                    ", busy computing a new image.", "black")

                    # Perform the new computation. Try to store intermediate results. If it fails
                    # because there is not enough RAM, switch to direct computation.
                    if self.ram_sufficient:
                        try:
                            self.postproc_data_object.versions[
                                self.version_selected].image = self.recompute_selected_version(
                                                                self.layers_selected)
                        except:
                            self.ram_sufficient = False
                            self.postproc_data_object.versions[
                                self.version_selected].image = Miscellaneous.post_process(
                                                            self.input_image, self.layers_selected)
                    else:
                        self.postproc_data_object.versions[
                            self.version_selected].image = Miscellaneous.post_process(
                            self.input_image, self.layers_selected)

                    # Reset the status bar to its idle state.
                    self.set_status_bar_signal.emit("Postprocessing " + self.file_name, "black")

                    # Show the new image in the image viewer, and remember its parameters.
                    self.set_photo_signal.emit(self.version_selected)
                    self.last_version_selected = self.version_selected

                # Neither the shift nor parameters have changed. If the version selection has changed,
                # display the image stored with the version selected.
                elif self.version_selected != self.last_version_selected:
                    # Show the new image in the image viewer, and remember its parameters.
                    self.set_photo_signal.emit(self.version_selected)
                    self.last_version_selected = self.version_selected

                # Idle loop before doing the next check for updates.
                else:
                    sleep(self.postproc_idle_loop_time)

                # Reset any exception condition after a successful loop pass.
                exception_status = False

            except Exception as e:
                # Print the error message only once. The error status is reset after the next
                # successful loop pass.
                if not exception_status:
                    print ("Exception in ImageProcessor: " + str(e) + ". Will try again.",
                           file=stderr)
                    exception_status = True
                sleep(self.postproc_idle_loop_time)

    def reset_intermediate_images(self):
        """
        Initialize all intermediate image versions, so that they will be re-computed next time.

        :return: -
        """

        self.layer_input = [None] * (self.configuration.postproc_max_layers + 1)
        self.layer_gauss = [None] * (self.configuration.postproc_max_layers)
        self.layer_bilateral = [None] * (self.configuration.postproc_max_layers)
        self.layer_denoised = [None] * (self.configuration.postproc_max_layers)

    def new_computation_required(self, version_has_changed):
        """
        Decide if a new computation is required. Invalidate only those intermediate image
        components which are affected by the parameter change. This results in a substantial
        performance improvement.

        :param version_has_changed: True, if the version has changed, so all intermediate results
                                    must be invalidated.
        :return: True, if the current version parameters have changed. False, otherwise.
        """

        # Initialize the flag which remembers if a change was found. Do not return on the first
        # detection in order to invalidate all affected intermediate results.
        change_detected = False

        # If a new version is selected, reset images with intermediate results.
        if version_has_changed:
            self.reset_intermediate_images()
            change_detected = True

        # Check if an additional version was created. Copy its layer info for later checks
        # for changes. The intermediate results are reset already because the version number must
        # have changed.
        if self.version_selected >= len(self.last_version_layers):
            self.last_version_layers.append(deepcopy(self.postproc_data_object.versions[
                                                         self.version_selected].layers))
            return change_detected

        # If the selected version is already known, check if the number of layers has changed.
        if len(self.last_version_layers[self.version_selected]) != len(self.layers_selected):
            self.last_version_layers[self.version_selected] = deepcopy(self.layers_selected)
            self.reset_intermediate_images()
            change_detected = True
            return change_detected

        # For color images, check if the "luminance only" parameter has changed. In this case all
        # layers are invalidated completely. Remember that the "luminance only" parameter is the
        # same on all layers for any given version, and that for version 0 (original image) no
        # layers are defined (and therefore there is no parameter "luminance_only").
        if self.color and self.version_selected and \
                self.last_version_layers[self.version_selected][0].luminance_only != \
                self.layers_selected[0].luminance_only:
            self.reset_intermediate_images()
            change_detected = True

        # The version is the same as before, it has the same number of layers, and the luminance
        # parameter has not changed. Now go through all layers and invalidate those intermediate
        # results individually which need to be recomputed.
        else:
            for layer_index, (last_layer, layer_selected) in enumerate(
                    zip(self.last_version_layers[self.version_selected], self.layers_selected)):
                # The "amount" parameter has changed. No intermediate results need to be invalidated
                # because the layer's contribution to the image is computed by multiplying the
                # denoised component with the amount parameter.
                if last_layer.amount != layer_selected.amount:
                    last_layer.amount = layer_selected.amount
                    change_detected = True

                # If the radius has changed, all filters at this level and the input for the next
                # level have to be recomputed.
                if  last_layer.radius != layer_selected.radius:
                    self.layer_gauss[layer_index] = self.layer_bilateral[layer_index] = \
                        self.layer_denoised[layer_index] = self.layer_input[layer_index + 1] = None
                    change_detected = True

                # If the bi_fraction has changed, the input for the next layer and the denoised
                # component on this layer must be recomputed.
                if last_layer.bi_fraction != layer_selected.bi_fraction:
                    self.layer_denoised[layer_index] = self.layer_input[layer_index + 1] = None
                    change_detected = True

                # If the bi_range parameter has changed, the bilateral filter is invalidated.
                if last_layer.bi_range != layer_selected.bi_range:
                    self.layer_bilateral[layer_index] = None
                    change_detected = True
                    # If the bi_fraction parameter is not zero, the change propagates to the
                    # denoised component and the input for the next level.
                    if abs(layer_selected.bi_fraction) > 1.e-5:
                        self.layer_denoised[layer_index] = self.layer_input[layer_index + 1] = None

                # If the denoise parameter has changed, the denoised component for this layer is
                # invalidated.
                if last_layer.denoise != layer_selected.denoise:
                    self.layer_denoised[layer_index] = None
                    change_detected = True

        # Remember the current parameter settings to compare with new parameters next time.
        self.last_version_layers[self.version_selected] = deepcopy(self.layers_selected)
        return change_detected

    def recompute_selected_version(self, layers):
        """
        Do the actual computation. Starting from the original image, for each sharpening layer
        apply a Gaussian filter using the layer's parameters. Store the result in the central
        data object for the selected version.

        :return: -
        """

        # Check if the original image is selected (version 0). In this case return the (potentially
        # RGB-shifted) original iamge.
        if not layers:
            return self.input_image.astype(uint16)

        for layer_index, layer in enumerate(layers):
            if self.layer_input[layer_index] is None:
                # For layers > 0, the layer_input must have been computed on the previous layer.
                if layer_index:
                    raise InternalError("Layer input image is None for layer > 0")
                # On layer 0, the original image is taken as layer input.
                else:
                    # For color input and luminance_only: extract the luminance channel.
                    if self.color and layer.luminance_only:
                        self.layer_input[layer_index] = self.input_image_hsv[:, :, 2]
                    else:
                        self.layer_input[layer_index] = self.input_image

            # Bilateral filter is needed:
            if abs(layer.bi_fraction) > 1.e-5:
                # Filter must be recomputed.
                if self.layer_bilateral[layer_index] is None:
                    self.layer_bilateral[layer_index] = bilateralFilter(self.layer_input[layer_index],
                        0, layer.bi_range * 256., layer.radius/3., borderType=BORDER_DEFAULT)

            # Gaussian filter is needed:
            if abs(layer.bi_fraction - 1.) > 1.e-5:
                # Filter must be recomputed.
                if self.layer_gauss[layer_index] is None:
                    self.layer_gauss[layer_index] = GaussianBlur(self.layer_input[layer_index], (0, 0),
                        layer.radius/3., borderType=BORDER_DEFAULT)

            # Compute the input image for the next layer.
            if self.layer_input[layer_index + 1] is None:
                # Case bilateral only.
                if abs(layer.bi_fraction - 1.) <= 1.e-5:
                    self.layer_input[layer_index + 1] = self.layer_bilateral[layer_index]
                # Case Gaussian only.
                elif abs(layer.bi_fraction) <= 1.e-5:
                    self.layer_input[layer_index + 1] = self.layer_gauss[layer_index]
                # Mixed case.
                else:
                    self.layer_input[layer_index + 1] = self.layer_bilateral[layer_index] * layer.bi_fraction + \
                                                        self.layer_gauss[layer_index] * (1. - layer.bi_fraction)

            # A new denoised layer component must be computed.
            if self.layer_denoised[layer_index] is None:
                layer_component_before_denoise = self.layer_input[layer_index] - self.layer_input[layer_index + 1]
                # Denoising must be applied.
                if layer.denoise > 1.e-5:
                    self.layer_denoised[layer_index] = GaussianBlur(layer_component_before_denoise, (0, 0),
                        layer.radius / 3., borderType=BORDER_DEFAULT) * layer.denoise + \
                        layer_component_before_denoise * (1. - layer.denoise)
                else:
                    self.layer_denoised[layer_index] = layer_component_before_denoise

        # Build the sharpened image as a weighted sum of the layer components. Start with the
        # maximally blurred layer input (start image of first layer beyond active layers).
        # A weight > 1 increases details at this level, a weight < 1 lowers their visibility.
        new_image = copy(self.layer_input[len(layers)])
        for layer, image_layer_component in zip(layers, self.layer_denoised):
            new_image += image_layer_component * layer.amount

        # In case of "luminance only", insert the new luminance channel into a copy of the original
        # image and change back to BGR.
        if self.color and layers[0].luminance_only:
            new_image_bgr = copy(self.input_image_hsv)
            new_image_bgr[:, :, 2] = new_image
            new_image = cvtColor(new_image_bgr, COLOR_HSV2BGR)

        # Clip pixels out of range and convert the processed image to 16bit unsigned int.
        return new_image.clip(min=0., max=65535.).astype(uint16)

    def stop(self):
        """
        Terminate the ImageProcessor.
        :return: -
        """

        self.terminate()


class PostprocEditorWidget(QtWidgets.QFrame, Ui_postproc_editor):
    """
    This widget implements a frame viewer together with control elements to control the
    postprocessing. Several postprocessing versions can be created and managed, each one using
    a variable number of sharpening layers. A blink comparator allows comparing any two versions.
    """

    def __init__(self, configuration, image_original, name_original, set_status_bar_callback,
                 signal_save_postprocessed_image):
        """

        :param configuration: Configuration object with parameters.
        :param image_original: Original image (16bit) holding the input for postprocessing.
        :param name_original: Path name of the original image.
        :param set_status_bar_callback: Call-back function to update the main GUI's status bar.
        :param signal_save_postprocessed_image: Signal to be issued when the postprocessing
                                                widget closes. None if no signal is to be issued.
        """

        super(PostprocEditorWidget, self).__init__()
        self.setupUi(self)

        self.configuration = configuration
        self.postproc_data_object = self.configuration.postproc_data_object
        self.postproc_data_object.set_postproc_input_image(image_original, name_original,
                                                self.configuration.global_parameters_image_format)
        self.signal_save_postprocessed_image = signal_save_postprocessed_image

        self.tabWidget_postproc_control.currentChanged.connect(self.tab_changed)

        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.pushButton_add_layer.clicked.connect(self.add_layer)

        rgb_resolutions = ['     1 Pixel ', '   0.5 Pixels', '  0.25 Pixels']
        self.max_rgb_index = self.max_auto_rgb_index()
        for resolution_index in range(self.max_rgb_index + 1):
            self.comboBox_resolution.addItem(rgb_resolutions[resolution_index])

        self.fgw_slider_value.valueChanged['int'].connect(self.fgw_changed)
        self.comboBox_resolution.currentIndexChanged.connect(self.rgb_resolution_changed)
        self.checkBox_automatic.stateChanged.connect(self.rgb_automatic_changed)
        self.pushButton_red_reset.clicked.connect(self.prreset_clicked)
        self.pushButton_red_up.clicked.connect(self.pru_clicked)
        self.pushButton_red_down.clicked.connect(self.prd_clicked)
        self.pushButton_red_left.clicked.connect(self.prl_clicked)
        self.pushButton_red_right.clicked.connect(self.prr_clicked)
        self.pushButton_blue_reset.clicked.connect(self.pbreset_clicked)
        self.pushButton_blue_up.clicked.connect(self.pbu_clicked)
        self.pushButton_blue_down.clicked.connect(self.pbd_clicked)
        self.pushButton_blue_left.clicked.connect(self.pbl_clicked)
        self.pushButton_blue_right.clicked.connect(self.pbr_clicked)

        # Enable RGB alignment for color input only.
        self.tab_rgb.setEnabled(self.postproc_data_object.color)

        # Initialize list of sharpening layer widgets, and set the maximal number of layers.
        self.sharpening_layer_widgets = []
        self.max_layers = self.configuration.postproc_max_layers

        self.label_message.setText("Edit postproc versions with up to "
                                   + str(self.max_layers) + " correction layers.")
        self.label_message.setStyleSheet('color: red')

        # Start the frame viewer.
        self.frame_viewer = FrameViewer()
        self.frame_viewer.setObjectName("framewiever")
        self.gridLayout.addWidget(self.frame_viewer, 0, 0, 2, 2)
        self.frame_viewer.zoom_factor_signal.connect(self.display_zoom_factor)

        # Initialize a vertical spacer used to fill the lower part of the sharpening widget scroll
        # area.
        self.spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum,
                                            QtWidgets.QSizePolicy.Expanding)

        # Set the resolution index to an impossible value. It is used to check for changes.
        self.rgb_resolution_index = -1

        # Set the vertical image size (in pixels) to an impossible value. It is used to check for
        # changes later. When the pixel resolution of the displayed image changes, the "fit in view"
        # method of the image viewer is invoked to normalize the appearance of the image.
        self.image_size_y = -1

        # Create the version manager.
        self.selected_version = None
        self.version_manager_widget = VersionManagerWidget(self.configuration)
        self.gridLayout.addWidget(self.version_manager_widget, 1, 2, 1, 1)

        # The "set_photo_signal" from the VersionManagerWidget is not passed to the image viewer
        # directly. (The image viewer does not accept signals.) Instead, it sends a signal to this
        # object which in turn invokes the image viewer.
        self.version_manager_widget.set_photo_signal.connect(self.select_image)
        self.version_manager_widget.select_version_signal.connect(self.select_version)

        # Select the initial current version.
        self.select_version(self.postproc_data_object.version_selected)

        # Start the image processor thread, and connect its signals.
        self.image_processor = ImageProcessor(self.configuration)
        self.image_processor.setTerminationEnabled(True)
        self.image_processor.set_photo_signal.connect(self.select_image)
        self.image_processor.set_shift_display_signal.connect(self.display_shifts)
        self.image_processor.set_status_bar_signal.connect(set_status_bar_callback)
        self.image_processor.disable_widgets_signal.connect(self.disable_widgets)
        self.image_processor.enable_widgets_signal.connect(self.enable_widgets)

        if self.max_rgb_index == -1:
            self.checkBox_automatic.setChecked(False)
            self.tab_rgb.setEnabled(False)

        for version in self.postproc_data_object.versions:
            if self.max_rgb_index == -1:
                version.rgb_automatic = False
            version.rgb_resolution_index = min(version.rgb_resolution_index, max(self.max_rgb_index, 0))

    def max_auto_rgb_index(self):
        """
        Return the maximal auto RGB alignment index such that the image created is not larger than
        10% of the available RAM. This makes sure that the postprocessing does not run out of RAM.

        :return: Maximal RGB alignment index (0, 1, or 2). If the image is too large to do any
                 auto RGB alignment, -1 is returned.
        """

        level = 2
        max_index = 2
        factors = [1, 2, 4]

        # Look up the available RAM (without paging)
        virtual_memory = dict(psutil.virtual_memory()._asdict())
        available_ram = virtual_memory['available']

        # For monochrome images, nothing is to be done.
        if self.postproc_data_object.color:
            image = self.postproc_data_object.image_original

            # The intermediate image is in float32 (4 bytes) and has 3 color channels.
            size_of_image = image.shape[0] * image.shape[1] * 12.

            # Compute the maximal interpolation factor.
            max_factor = int(sqrt(
                available_ram * self.configuration.postproc_max_ram_percentage / 100. / size_of_image))

            # In this case auto RGB alignment does not even work for standard size.
            if max_factor == 0:
                return -1

            # Look for the maximal interpolation index.
            for level in range(max_index, -1, -1):
                if factors[level] <= max_factor:
                    break

        return level

    def disable_widgets(self):
        """
        To avoid race conditions, GUI widgets are disabled while computations are going on. The only
        button remaining active is "Cancel".

        :return: -
        """

        self.version_manager_widget.setEnabled(False)
        self.tabWidget_postproc_control.setEnabled(False)
        self.buttonBox.button(QtWidgets.QDialogButtonBox.Ok).setEnabled(False)

    def enable_widgets(self):
        """
        Enable the GUI buttons when background computations have finished.

        :return: -
        """

        self.version_manager_widget.setEnabled(True)
        self.tabWidget_postproc_control.setEnabled(True)
        self.buttonBox.button(QtWidgets.QDialogButtonBox.Ok).setEnabled(True)

    def fgw_changed(self, value):
        """
        When the widget changes its value, update the corresponding entry for the current version.
        Please note that for some parameters the representations differ.

        The methods following this one do the same for all other configuration parameters.

        :param value: New value sent by widget
        :return: -
        """

        gauss_width = 2 * value - 1
        self.postproc_data_object.versions[self.postproc_data_object.version_selected].rgb_gauss_width = gauss_width
        self.fgw_label_display.setText(str(gauss_width))

    def rgb_resolution_changed(self, value):
        version = self.postproc_data_object.versions[self.postproc_data_object.version_selected]
        version.rgb_resolution_index = value

        # Round the shift values according to the new resolution.
        factor = [1., 2., 4.][value]
        factor_incremented = [1.01, 2.01, 4.01][value]
        version.shift_red = (round(version.shift_red[0] * factor_incremented) / factor ,
                             round(version.shift_red[1] * factor_incremented) / factor)
        version.shift_blue = (round(version.shift_blue[0] * factor_incremented) / factor,
                              round(version.shift_blue[1] * factor_incremented) / factor)
        version.correction_red = (round(version.correction_red[0] * factor_incremented) / factor,
                             round(version.correction_red[1] * factor_incremented) / factor)
        version.correction_blue = (round(version.correction_blue[0] * factor_incremented) / factor,
                              round(version.correction_blue[1] * factor_incremented) / factor)
        self.display_shifts()

    def tab_changed(self, index):
        version = self.postproc_data_object.versions[
            self.postproc_data_object.version_selected]
        if index and not version.rgb_automatic:
            self.rgb_correction_version_init(version)
        elif not index:
            # Force the computation of a new image in the ImageProcessor.
            version.shift_red_saved = None
            version.shift_blue_saved = None
            self.finish_rgb_correction_mode()

    def rgb_automatic_changed(self, state):
        rgb_on = state == QtCore.Qt.Checked
        version = self.postproc_data_object.versions[
            self.postproc_data_object.version_selected]
        version.rgb_automatic = rgb_on
        if rgb_on:
            self.rgb_correction_version_reset(version)
        else:
            self.rgb_correction_version_init(version)

    def rgb_correction_version_init(self, version):
        version.rgb_correction_mode = True
        version.correction_red_saved = version.correction_blue_saved = version.correction_red = \
            version.correction_blue = (0., 0.)
        version.images_uncorrected = [None] * 3

    def rgb_correction_version_reset(self, version):
        version.rgb_correction_mode = False
        version.correction_red_saved = version.correction_blue_saved = version.correction_red =\
            version.correction_blue = (0., 0.)
        version.images_uncorrected = [None] * 3

    def finish_rgb_correction_mode(self):
        for version_index, version in enumerate(self.postproc_data_object.versions):
            if version.correction_red != (0., 0.) or version.correction_blue != (0., 0.):
                version.shift_red = (version.shift_red[0] + version.correction_red[0],
                                     version.shift_red[1] + version.correction_red[1])
                version.shift_blue = (version.shift_blue[0] + version.correction_blue[0],
                                      version.shift_blue[1] + version.correction_blue[1])
            self.rgb_correction_version_reset(version)
        self.select_image(self.postproc_data_object.version_selected)

    def prreset_clicked(self):
        version = self.postproc_data_object.versions[self.postproc_data_object.version_selected]
        version.correction_red = (-version.shift_red[0], -version.shift_red[1])
        self.display_shifts()

    def pru_clicked(self):
        version = self.postproc_data_object.versions[self.postproc_data_object.version_selected]
        increment = [1., 0.5, 0.25][version.rgb_resolution_index]
        version.correction_red = (version.correction_red[0] - increment, version.correction_red[1])
        self.display_shifts()

    def prd_clicked(self):
        version = self.postproc_data_object.versions[self.postproc_data_object.version_selected]
        increment = [1., 0.5, 0.25][version.rgb_resolution_index]
        version.correction_red = (version.correction_red[0] + increment, version.correction_red[1])
        self.display_shifts()

    def prl_clicked(self):
        version = self.postproc_data_object.versions[self.postproc_data_object.version_selected]
        increment = [1., 0.5, 0.25][version.rgb_resolution_index]
        version.correction_red = (version.correction_red[0], version.correction_red[1] - increment)
        self.display_shifts()

    def prr_clicked(self):
        version = self.postproc_data_object.versions[self.postproc_data_object.version_selected]
        increment = [1., 0.5, 0.25][version.rgb_resolution_index]
        version.correction_red = (version.correction_red[0], version.correction_red[1] + increment)
        self.display_shifts()

    def pbreset_clicked(self):
        version = self.postproc_data_object.versions[self.postproc_data_object.version_selected]
        version.correction_blue = (-version.shift_blue[0], -version.shift_blue[1])
        self.display_shifts()

    def pbu_clicked(self):
        version = self.postproc_data_object.versions[self.postproc_data_object.version_selected]
        increment = [1., 0.5, 0.25][version.rgb_resolution_index]
        version.correction_blue = (version.correction_blue[0] - increment, version.correction_blue[1])
        self.display_shifts()

    def pbd_clicked(self):
        version = self.postproc_data_object.versions[self.postproc_data_object.version_selected]
        increment = [1., 0.5, 0.25][version.rgb_resolution_index]
        version.correction_blue = (
        version.correction_blue[0] + increment, version.correction_blue[1])
        self.display_shifts()

    def pbl_clicked(self):
        version = self.postproc_data_object.versions[self.postproc_data_object.version_selected]
        increment = [1., 0.5, 0.25][version.rgb_resolution_index]
        version.correction_blue = (version.correction_blue[0], version.correction_blue[1] - increment)
        self.display_shifts()

    def pbr_clicked(self):
        version = self.postproc_data_object.versions[self.postproc_data_object.version_selected]
        increment = [1., 0.5, 0.25][version.rgb_resolution_index]
        version.correction_blue = (
        version.correction_blue[0], version.correction_blue[1] + increment)
        self.display_shifts()

    def display_shifts(self):
        """
        Set the current channel shifts in the RGB alignment GUI tab.

        :return: -
        """

        format_string = ["{0:2.0f}", "{0:4.1f}", "{0:5.2f}"][self.postproc_data_object.versions[
            self.postproc_data_object.version_selected].rgb_resolution_index]

        # Red channel shifts:
        shift_red_base = self.postproc_data_object.versions[
            self.postproc_data_object.version_selected].shift_red
        correction_red = self.postproc_data_object.versions[
            self.postproc_data_object.version_selected].correction_red
        shift_red = (shift_red_base[0] + correction_red[0], shift_red_base[1] + correction_red[1])

        if abs(shift_red[0]) < 0.05:
            self.label_red_down.setText("")
            self.label_red_up.setText("")
        elif shift_red[0] > 0:
            self.label_red_down.setText((format_string.format(shift_red[0])))
            self.label_red_up.setText("")
        else:
            self.label_red_up.setText((format_string.format(-shift_red[0])))
            self.label_red_down.setText("")

        if abs(shift_red[1]) < 0.05:
            self.label_red_left.setText("")
            self.label_red_right.setText("")
        elif shift_red[1] > 0:
            self.label_red_right.setText((format_string.format(shift_red[1])))
            self.label_red_left.setText("")
        else:
            self.label_red_left.setText((format_string.format(-shift_red[1])))
            self.label_red_right.setText("")

        # Blue channel shifts:
        shift_blue_base = self.postproc_data_object.versions[
            self.postproc_data_object.version_selected].shift_blue
        correction_blue = self.postproc_data_object.versions[
            self.postproc_data_object.version_selected].correction_blue
        shift_blue = (shift_blue_base[0] + correction_blue[0], shift_blue_base[1] + correction_blue[1])

        if abs(shift_blue[0]) < 0.05:
            self.label_blue_down.setText("")
            self.label_blue_up.setText("")
        elif shift_blue[0] > 0:
            self.label_blue_down.setText((format_string.format(shift_blue[0])))
            self.label_blue_up.setText("")
        else:
            self.label_blue_up.setText((format_string.format(-shift_blue[0])))
            self.label_blue_down.setText("")

        if abs(shift_blue[1]) < 0.05:
            self.label_blue_left.setText("")
            self.label_blue_right.setText("")
        elif shift_blue[1] > 0:
            self.label_blue_right.setText((format_string.format(shift_blue[1])))
            self.label_blue_left.setText("")
        else:
            self.label_blue_left.setText((format_string.format(-shift_blue[1])))
            self.label_blue_right.setText("")

    def display_zoom_factor(self, zoom_factor):
        """
        Write the current viewer zoom factor under the viewer window.

        :param zoom_factor: Current zoom factor, sent by the FrameViewer via a signal.

        :return: -
        """

        self.label_zoom.setText("Zoom: " + str(zoom_factor) + "%")

    def select_version(self, version_index):
        """
        Select a new current version, update the scroll area with all layer widgets, and update the
        parameters of the RGB alignment tab.

        :param version_index: Index of the version selected in version list of the central data
                              object.
        :return: -
        """

        # Check if we are leaving a version in correction mode. In this case apply the corrections.
        if self.selected_version is not None:
            if self.selected_version.rgb_correction_mode:
                self.finish_rgb_correction_mode()

        self.selected_version = self.postproc_data_object.versions[version_index]

        # Remove all existing layer widgets and the lower vertical spacer.
        if self.sharpening_layer_widgets:
            for layer_widget in self.sharpening_layer_widgets:
                self.verticalLayout.removeWidget(layer_widget)
                layer_widget.deleteLater()
                del layer_widget
        self.sharpening_layer_widgets = []
        self.verticalLayout.removeItem(self.spacerItem)

        # Create new layer widgets for all active layers and put them into the scroll area.
        for layer_index, layer in enumerate(self.selected_version.layers):
            sharpening_layer_widget = SharpeningLayerWidget(layer_index, self.remove_layer)
            sharpening_layer_widget.set_values(layer)
            self.verticalLayout.addWidget(sharpening_layer_widget)
            self.sharpening_layer_widgets.append(sharpening_layer_widget)

        # Connect the signal issued by a sharpening layer when the "luminance only" checkbox changes
        # state with all other layers, so they do the same.
        for layer_index, sharpening_layer_widget in enumerate(self.sharpening_layer_widgets):
            toggle_luminance_signal = sharpening_layer_widget.signal_toggle_luminance_other_layers
            for other_layer_index, other_sharpening_layer_widget in enumerate(
                    self.sharpening_layer_widgets):
                if layer_index != other_layer_index:
                    toggle_luminance_signal.connect(
                        other_sharpening_layer_widget.checkBox_luminance_toggled_no_signal)

        # At the end of the scroll area, add a vertical spacer.
        self.verticalLayout.addItem(self.spacerItem)

        # Load the parameters of this version into the RGB alignment tab. Block signals to avoid
        # triggering change actions which are not required at this point.
        self.checkBox_automatic.blockSignals(True)
        self.comboBox_resolution.blockSignals(True)
        self.checkBox_automatic.setChecked(self.selected_version.rgb_automatic)
        self.comboBox_resolution.setCurrentIndex(min(self.selected_version.rgb_resolution_index,
                                                     self.max_rgb_index))
        self.checkBox_automatic.blockSignals(False)
        self.comboBox_resolution.blockSignals(False)
        self.fgw_slider_value.setValue(int((self.selected_version.rgb_gauss_width + 1) / 2))
        self.fgw_label_display.setText(str(self.selected_version.rgb_gauss_width))

        # Only if the RGB tab is selected: Enter / leave RGB correction mode.
        if self.tabWidget_postproc_control.currentIndex():
            if self.selected_version.rgb_automatic:
                self.rgb_correction_version_reset(self.selected_version)
            else:
                self.rgb_correction_version_init(self.selected_version)

        # Load the current image into the image viewer.
        self.select_image(version_index)

        # Display the current RGB shifts for the selected version.
        self.display_shifts()

    def select_image(self, version_index):
        """
        Load the current image into the image viewer. This method is either invoked by a direct
        call or by a "set_photo_signal" signal.

        :param version_index: Index of the version selected in version list of the central data
                              object.
        :return: -
        """

        image = self.postproc_data_object.versions[version_index].image
        self.frame_viewer.setPhoto(image)

        # Check if the image size has changed by more than 10%. If so, reset the zoom factor of the
        # FrameViewer.
        # print ("Select image")
        if abs((image.shape[0] - self.image_size_y)/ image.shape[0]) > 0.1:
            # print("Fit in view")
            self.frame_viewer.fitInView()
            self.image_size_y = image.shape[0]

    def add_layer(self):
        """
        This method is called when the "Add postprocessing layer" button is pressed. Add a layer
        to the currently selected postprocessing version.

        :return:-
        """

        version_selected = self.postproc_data_object.versions[
            self.postproc_data_object.version_selected]
        num_layers_current = version_selected.number_layers

        # The "zero" layer is reserved for the original image, so for this version no additional
        # layer can be allocated.

        # The first layer is initialized with a fixed parameter set (see "init" method of DataObject
        # class).
        if self.postproc_data_object.version_selected and num_layers_current < self.max_layers:

            # This version has already at least one layer. Initialize the radius parameter of the
            # new layer to 1.5 times the radius of the previous one.
            if num_layers_current:
                previous_layer = version_selected.layers[num_layers_current - 1]
                new_layer = PostprocLayer(previous_layer.postproc_method,
                                          round(1.5 * previous_layer.radius, 1), 1.,
                                          previous_layer.bi_fraction,
                                          previous_layer.bi_range, 0.,
                                          previous_layer.luminance_only)

            # This is the first layer for this image version. Start with standard parameters.
            else:
                new_layer = PostprocLayer("Multilevel unsharp masking", 1., 1., 0., 20, 0., False)
            version_selected.add_postproc_layer(new_layer)

            # Update all layer widgets.
            self.select_version(self.postproc_data_object.version_selected)

    def remove_layer(self, layer_index):
        """
        Remove a layer from the selected version. The "remove_layer" method of class "Layer" makes
        sure that the first layer is not removed.

        :param layer_index: Index of the layer to be removed.
        :return: -
        """

        version_selected = self.postproc_data_object.versions[
            self.postproc_data_object.version_selected]
        version_selected.remove_postproc_layer(layer_index)

        # Update all layer widgets.
        self.select_version(self.postproc_data_object.version_selected)

    def accept(self):
        """
        When the user presses the "OK" button, save the currently selected version to the standard
        path and exit.

        :return: -
        """

        # In case "correction mode" is on, process the current version image in full 16bit
        # resolution.
        self.disable_widgets()
        self.finish_rgb_correction_mode()
        self.postproc_data_object.finalize_postproc_version()

        # Terminate the image processor thread.
        self.image_processor.stop()
        self.configuration.write_config()
        if self.signal_save_postprocessed_image:
            self.signal_save_postprocessed_image.emit(self.postproc_data_object.versions[
                self.postproc_data_object.version_selected].image)
        # The else branch is only for the test program below which does not provide a signal for
        # saving the result. In this case use the save method of the version manager.
        else:
            self.version_manager_widget.save_version()
        self.enable_widgets()
        self.close()

    def reject(self):
        """
        When the user presses the "Cancel" button, terminate the image processor thread and
        exit without saving anything.

        :return: -
        """

        self.image_processor.stop()
        if self.signal_save_postprocessed_image:
            self.signal_save_postprocessed_image.emit(None)
        self.close()


class EmulateStatusBar(object):
    """
    This class is only used for unit testing with the main program below. It emulates the status
    bar of the main GUI by printing the message to standard output.
    """

    def __init__(self):
        super(EmulateStatusBar, self).__init__()
        pass

    @staticmethod
    def print_status_bar_info(message, color):
        """
        Emulate the "write_status_bar" method of the main GUI.

        :param message: Message to be displayed (str).
        :param color: String with the color of the message (e.g. "red").
        :return:
        """

        colors = {}
        colors['red'] = "\033[1;31m"
        colors['blue'] = "\033[1;34m"
        colors['cyan'] = "\033[1;36m"
        colors['green'] = "\033[0;32m"
        colors['reset'] = "\033[0;0m"
        colors['bold'] = "\033[;1m"

        # Add a try-except clause to avoid a program crash if an un-supported color is specified.
        # In this case, just use a standard print.
        try:
            stdout.write(colors[color])
            print(message)
            stdout.write(colors['reset'])
        except:
            print(message)


if __name__ == '__main__':
    # input_file_name = "D:\SW-Development\Python\PlanetarySystemStacker\Examples\Moon_2018-03-24\Moon_Tile-024_043939_pss.tiff"
    input_file_name = "D:\SW-Development\Python\PlanetarySystemStacker\Examples\Jupiter_Richard\\" \
                      "2020-07-29-2145_3-L-Jupiter_ALTAIRGP224C_pss_p70_b48.png"
    # input_file_name = "D:\SW-Development\Python\PlanetarySystemStacker\Examples\Jupiter\\2019-05-26-0115_4-L-Jupiter_ZWO ASI290MM Mini_pipp_pss.png"
    # input_file_name = "D:\SW-Development\Python\PlanetarySystemStacker\Examples\Moon_2011-04-10\South.stacked.tiff"

    input_image = imread(input_file_name, -1)
    # Change colors to standard RGB
    if len(input_image.shape) == 3:
        input_image = cvtColor(input_image, COLOR_BGR2RGB)

    configuration = Configuration()
    configuration.initialize_configuration()
    dummy_status_bar = EmulateStatusBar()
    app = QtWidgets.QApplication(argv)
    window = PostprocEditorWidget(configuration, input_image, input_file_name,
                                  dummy_status_bar.print_status_bar_info, None)
    window.show()
    app.exec_()



from math import ceil
from time import time

from cv2 import resize, INTER_AREA, INTER_LINEAR, INTER_CUBIC, imshow, waitKey, destroyAllWindows, \
    imread, IMREAD_GRAYSCALE, namedWindow, setWindowProperty, WND_PROP_FULLSCREEN, \
    WINDOW_FULLSCREEN, GaussianBlur
from numpy import arange, zeros, float32

from miscellaneous import Miscellaneous


def drizzle_shift():
    shifts = arange(-3., 3., 0.1, dtype=float)
    for shift in shifts:
        shift_ceil = ceil(shift / 2) * 2
        frac = shift_ceil - shift - 1. / 3.
        frac_shift = int(round(frac * 3. / 2.))

        print("shift: " + str(shift) + ", ceil: " + str(shift_ceil) + ", frac:" + str(frac_shift))


def drizzle_shift_patch(shift, ilo, ihi):
    frame = arange(30.)
    frame_d = zeros((45,))

    shift_ceil_even = ceil((shift - 1. / 3.) / 2) * 2
    frac = shift_ceil_even - shift
    frac_shift_even = int(round(frac * 3. / 2.))
    shift_ceil_odd = ceil(shift / 2) * 2
    frac = shift_ceil_odd - shift - 1. / 3.
    frac_shift_odd = int(round(frac * 3. / 2.))
    if ilo % 2:
        frame_low_even = ilo + 1 + shift_ceil_even
        frame_high_even = ihi + shift_ceil_even
        frame_d_low_even = int((ilo + 1) / 2 * 3 + frac_shift_even)
        frame_d_high_even = int(round(ihi / 2 * 3)) + frac_shift_even
        frame_low_odd = ilo + shift_ceil_odd
        frame_high_odd = ihi + shift_ceil_odd
        frame_d_low_odd = int((ilo + 1. / 3.) / 2 * 3 + frac_shift_odd)
        frame_d_high_odd = int(round(ihi / 2 * 3)) + frac_shift_odd
    else:
        frame_low_even = ilo + shift_ceil_even
        frame_high_even = ihi + shift_ceil_even
        frame_d_low_even = int(ilo / 2 * 3 + frac_shift_even)
        frame_d_high_even = int(ihi / 2 * 3 + frac_shift_even)
        frame_low_odd = ilo + 1 + shift_ceil_odd
        frame_high_odd = ihi + shift_ceil_odd
        frame_d_low_odd = int((ilo + 4. / 3.) / 2 * 3 + frac_shift_odd)
        frame_d_high_odd = int(ihi / 2 * 3 + frac_shift_odd)
    frame_d[frame_d_low_even:frame_d_high_even:3] = frame[frame_low_even:frame_high_even:2]
    frame_d[frame_d_low_odd:frame_d_high_odd:3] = frame[frame_low_odd:frame_high_odd:2]

    pass


def resize_test(image, drizzle_factor, inter, patch_size):
    width = int(image.shape[1] * drizzle_factor)
    height = int(image.shape[0] * drizzle_factor)
    dim = (width, height)
    time_before = time()
    rep_count = 100
    for i in range(rep_count):
        resized = resize(image, dim, interpolation=inter)
    exec_time = (time() - time_before) / rep_count
    print("Time for one resize operation: " + str(exec_time) + ", method: " +
          str(["INTER_LINEAR", "INTER_CUBIC", "INTER_AREA"][inter - 1]))
    # imshow("Resized image", resized)
    # waitKey(0)
    # destroyAllWindows()

    start_y = 350
    start_x = 500
    patch = image[start_y:start_y + patch_size, start_x:start_x + patch_size]
    width = int(patch.shape[1] * drizzle_factor)
    height = int(patch.shape[0] * drizzle_factor)
    dim = (width, height)
    rep_count_patch = int(
        round(rep_count * image.shape[1] * image.shape[0] / (patch.shape[1] * patch.shape[0])))
    time_before = time()
    for i in range(rep_count_patch):
        resized = resize(patch, dim, interpolation=inter)
    exec_time_patched = (time() - time_before) / rep_count
    factor = exec_time_patched / exec_time
    print("Time for one patched resize operation: " + str(exec_time_patched) + ", factor: "
          + str(factor) + str("\n"))


def show_image(window_name, image, fullscreen=False):
    if fullscreen:
        namedWindow(window_name, WND_PROP_FULLSCREEN)
        setWindowProperty(window_name, WND_PROP_FULLSCREEN, WINDOW_FULLSCREEN)
    imshow(window_name, image)
    waitKey()
    destroyAllWindows()


def subpixel_shifted_frame(frame, subpixel_shift_y, subpixel_shift_x):
    interpolation_method = INTER_CUBIC
    undersampling_factor = 0.25
    interpol_factor = 10
    width = int(frame.shape[1] * interpol_factor)
    height = int(frame.shape[0] * interpol_factor)
    dim = (width, height)
    frame_extended = resize(frame, dim, interpolation=interpolation_method)
    print("shape original: " + str(frame.shape) + ", shape extended: " + str(
        frame_extended.shape))
    shift_factor = interpol_factor / undersampling_factor
    shift_y = int(round(subpixel_shift_y * shift_factor))
    shift_x = int(round(subpixel_shift_x * shift_factor))
    border_y = abs(shift_y)
    border_x = abs(shift_x)

    frame_cropped_original = frame_extended[border_y:frame_extended.shape[0] - border_y,
                             border_x:frame_extended.shape[1] - border_x]
    frame_cropped_shifted = frame_extended[
                            border_y + shift_y:frame_extended.shape[0] - border_y + shift_y,
                            border_x + shift_x:frame_extended.shape[1] - border_x + shift_x]

    width = int(round(frame_cropped_original.shape[1] / shift_factor))
    height = int(round(frame_cropped_original.shape[0] / shift_factor))
    dim = (width, height)

    frame_resized = resize(frame_cropped_original, dim, interpolation=interpolation_method)
    frame_shifted = resize(frame_cropped_shifted, dim, interpolation=interpolation_method)

    print("shape resized: " + str(frame_resized.shape) + ", shape shifted: " + str(
        frame_shifted.shape))
    return frame_resized, frame_shifted


def drizzle_shift_test():
    drizzle_shift()
    drizzle_shift_patch(-1.4, 2, 7)

def interpolation_timing():
    img = imread('../Images/Moon_Tile-024_043939_stacked_with_blurr_pp.tif', IMREAD_GRAYSCALE)
    # show_image("Original image", img, fullscreen=True)

    drizzle_factor = 3.
    patch_size = 40

    for inter in [INTER_AREA, INTER_LINEAR, INTER_CUBIC]:
        resize_test(img, drizzle_factor, inter, patch_size)

def sub_pixel_shift_test():
    img = imread('../Images/Moon_Tile-024_043939_stacked_with_blurr_pp.tif', IMREAD_GRAYSCALE)
    show_image("Original image", img, fullscreen=True)
    spx_shifty = 5.2
    spx_shiftx = 3.5

    img_resized, img_shifted = subpixel_shifted_frame(img, spx_shifty, spx_shiftx)
    # for i in range(10):
    #     show_image("Image resized", img_resized, fullscreen=True)
    #     show_image("Image shifted", img_shifted, fullscreen=True)

    gauss_width_reference = 15
    gauss_width_frame = 19
    reference_frame_blurred_intermediate = GaussianBlur(img_resized,
                                                        (gauss_width_reference, gauss_width_reference),
                                                        0).astype(float32)
    reference_frame_blurred = GaussianBlur(reference_frame_blurred_intermediate,
                                           (gauss_width_reference, gauss_width_reference), 0).astype(
                                            float32)
    frame_blurred = GaussianBlur(img_shifted, (gauss_width_frame, gauss_width_frame), 0)

    y_ap = 170
    x_ap = 200
    half_box_width = 24
    y_low = y_ap - half_box_width
    y_high = y_ap + half_box_width
    x_low = x_ap - half_box_width
    x_high = x_ap + half_box_width
    reference_box_second_phase = reference_frame_blurred[y_low: y_high, x_low: x_high]
    reference_box_first_phase = reference_box_second_phase[::2, ::2]

    search_width = 10
    shift_y_local_first_phase, shift_x_local_first_phase, success_first_phase, \
    shift_y_local_second_phase, shift_x_local_second_phase, success_second_phase = \
        Miscellaneous.multilevel_correlation(reference_box_first_phase, frame_blurred,
                                             gauss_width_frame,
                                             reference_box_second_phase, y_low, y_high, x_low, x_high,
                                             search_width,
                                             weight_matrix_first_phase=None, subpixel_solve=True)

    print("Shift in y, first phase: " + str(shift_y_local_first_phase) + ", second phase: " + str(
        shift_y_local_second_phase) + ", total: " + str(
        shift_y_local_first_phase + shift_y_local_second_phase))
    print("Shift in x, first phase: " + str(shift_x_local_first_phase) + ", second phase: " + str(
        shift_x_local_second_phase) + ", total: " + str(
        shift_x_local_first_phase + shift_x_local_second_phase))

# drizzle_shift_test()
interpolation_timing()
# sub_pixel_shift_test()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'frame_viewer_gui.ui'
#
# Created by: PyQt5 UI code generator 5.11.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_frame_viewer(object):
    def setupUi(self, frame_viewer):
        frame_viewer.setObjectName("frame_viewer")
        frame_viewer.resize(900, 630)
        frame_viewer.setLocale(QtCore.QLocale(QtCore.QLocale.English, QtCore.QLocale.UnitedStates))
        frame_viewer.setFrameShape(QtWidgets.QFrame.Panel)
        frame_viewer.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.grid_layout = QtWidgets.QGridLayout(frame_viewer)
        self.grid_layout.setObjectName("grid_layout")
        self.groupBox_frame_sorting = QtWidgets.QGroupBox(frame_viewer)
        self.groupBox_frame_sorting.setObjectName("groupBox_frame_sorting")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.groupBox_frame_sorting)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.radioButton_quality = QtWidgets.QRadioButton(self.groupBox_frame_sorting)
        self.radioButton_quality.setObjectName("radioButton_quality")
        self.gridLayout_2.addWidget(self.radioButton_quality, 1, 0, 1, 1)
        self.radioButton_chronological = QtWidgets.QRadioButton(self.groupBox_frame_sorting)
        self.radioButton_chronological.setObjectName("radioButton_chronological")
        self.gridLayout_2.addWidget(self.radioButton_chronological, 0, 0, 1, 1)
        self.spinBox_chronological = QtWidgets.QSpinBox(self.groupBox_frame_sorting)
        self.spinBox_chronological.setObjectName("spinBox_chronological")
        self.gridLayout_2.addWidget(self.spinBox_chronological, 0, 1, 1, 1)
        self.spinBox_quality = QtWidgets.QSpinBox(self.groupBox_frame_sorting)
        self.spinBox_quality.setObjectName("spinBox_quality")
        self.gridLayout_2.addWidget(self.spinBox_quality, 1, 1, 1, 1)
        self.grid_layout.addWidget(self.groupBox_frame_sorting, 3, 3, 1, 1)
        self.groupBox_stacking_fraction = QtWidgets.QGroupBox(frame_viewer)
        self.groupBox_stacking_fraction.setObjectName("groupBox_stacking_fraction")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.groupBox_stacking_fraction)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.label_number_frames = QtWidgets.QLabel(self.groupBox_stacking_fraction)
        self.label_number_frames.setObjectName("label_number_frames")
        self.gridLayout_3.addWidget(self.label_number_frames, 0, 0, 1, 1)
        self.spinBox_number_frames = QtWidgets.QSpinBox(self.groupBox_stacking_fraction)
        self.spinBox_number_frames.setMinimum(1)
        self.spinBox_number_frames.setObjectName("spinBox_number_frames")
        self.gridLayout_3.addWidget(self.spinBox_number_frames, 0, 1, 1, 1)
        self.label_percentage_frames = QtWidgets.QLabel(self.groupBox_stacking_fraction)
        self.label_percentage_frames.setObjectName("label_percentage_frames")
        self.gridLayout_3.addWidget(self.label_percentage_frames, 1, 0, 1, 1)
        self.spinBox_percentage_frames = QtWidgets.QSpinBox(self.groupBox_stacking_fraction)
        self.spinBox_percentage_frames.setMinimum(1)
        self.spinBox_percentage_frames.setMaximum(100)
        self.spinBox_percentage_frames.setObjectName("spinBox_percentage_frames")
        self.gridLayout_3.addWidget(self.spinBox_percentage_frames, 1, 1, 1, 1)
        self.pushButton_set_stacking_limit = QtWidgets.QPushButton(self.groupBox_stacking_fraction)
        self.pushButton_set_stacking_limit.setMinimumSize(QtCore.QSize(250, 0))
        self.pushButton_set_stacking_limit.setObjectName("pushButton_set_stacking_limit")
        self.gridLayout_3.addWidget(self.pushButton_set_stacking_limit, 2, 0, 1, 2)
        self.grid_layout.addWidget(self.groupBox_stacking_fraction, 2, 3, 1, 1)
        self.buttonBox = QtWidgets.QDialogButtonBox(frame_viewer)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.grid_layout.addWidget(self.buttonBox, 4, 3, 1, 1)
        self.pushButton_play = QtWidgets.QPushButton(frame_viewer)
        self.pushButton_play.setObjectName("pushButton_play")
        self.grid_layout.addWidget(self.pushButton_play, 4, 2, 1, 1)
        self.pushButton_stop = QtWidgets.QPushButton(frame_viewer)
        self.pushButton_stop.setObjectName("pushButton_stop")
        self.grid_layout.addWidget(self.pushButton_stop, 4, 1, 1, 1)
        self.slider_frames = QtWidgets.QSlider(frame_viewer)
        self.slider_frames.setMaximum(1000)
        self.slider_frames.setPageStep(20)
        self.slider_frames.setOrientation(QtCore.Qt.Horizontal)
        self.slider_frames.setObjectName("slider_frames")
        self.grid_layout.addWidget(self.slider_frames, 4, 0, 1, 1)
        self.label_matplotlib = QtWidgets.QLabel(frame_viewer)
        self.label_matplotlib.setObjectName("label_matplotlib")
        self.grid_layout.addWidget(self.label_matplotlib, 0, 3, 2, 1)
        self.grid_layout.setColumnStretch(0, 5)

        self.retranslateUi(frame_viewer)
        QtCore.QMetaObject.connectSlotsByName(frame_viewer)

    def retranslateUi(self, frame_viewer):
        _translate = QtCore.QCoreApplication.translate
        frame_viewer.setWindowTitle(_translate("frame_viewer", "Frame"))
        self.groupBox_frame_sorting.setTitle(_translate("frame_viewer", "Frame sorting"))
        self.radioButton_quality.setToolTip(_translate("frame_viewer", "Frames are ordered by their overall sharpness."))
        self.radioButton_quality.setText(_translate("frame_viewer", "By quality"))
        self.radioButton_chronological.setToolTip(_translate("frame_viewer", "Frames are ordered chronologically."))
        self.radioButton_chronological.setText(_translate("frame_viewer", "Chronological"))
        self.spinBox_chronological.setToolTip(_translate("frame_viewer", "Enter a value or use the spin box arrows to select a (chronological) frame index."))
        self.spinBox_quality.setToolTip(_translate("frame_viewer", "Enter a value or use the spin box arrows to select a (quality) frame index."))
        self.groupBox_stacking_fraction.setTitle(_translate("frame_viewer", "Fraction of frames to be stacked"))
        self.label_number_frames.setText(_translate("frame_viewer", "Number of frames"))
        self.spinBox_number_frames.setToolTip(_translate("frame_viewer", "Enter a value or use the spin box arrows to set the number of frames to be stacked."))
        self.label_percentage_frames.setText(_translate("frame_viewer", "Percentage"))
        self.spinBox_percentage_frames.setToolTip(_translate("frame_viewer", "Enter a value or use the spin box arrows to set the percentage of frames to be stacked."))
        self.pushButton_set_stacking_limit.setToolTip(_translate("frame_viewer", "Adjust the number of best frames to be stacked at each alignment point such that the current frame is the worst one to be included."))
        self.pushButton_set_stacking_limit.setText(_translate("frame_viewer", "Set limit to current frame"))
        self.buttonBox.setToolTip(_translate("frame_viewer", "Exit the viewer. Press \'OK\' to save the stack size, or \'cancel\' to discard changes."))
        self.pushButton_play.setToolTip(_translate("frame_viewer", "Start a frame display video. Frames are ordered as selected in the \'frame sorting\' section."))
        self.pushButton_play.setText(_translate("frame_viewer", "Play"))
        self.pushButton_stop.setToolTip(_translate("frame_viewer", "Stop the frame display video."))
        self.pushButton_stop.setText(_translate("frame_viewer", "Stop"))
        self.slider_frames.setToolTip(_translate("frame_viewer", "Use the slider to select the frame to be displayed. As an alternative,\n"
"you can select the frame with the \'frame sorting\' spinboxes."))
        self.label_matplotlib.setText(_translate("frame_viewer", "Matplotlib placeholder"))




from numpy import ndarray
from numpy import uint8, uint16

from frames import Frames

image = ndarray((500, 500), dtype=uint8)
image[:,:] = 128
color = False
file_name = 'D:\SW-Development\Python\PlanetarySystemStacker\planetary_system_stacker\Images/new_file.tiff'
avoid_overwriting = True

# Save the stacked image as 16bit int.
Frames.save_image(file_name, image, color=color, avoid_overwriting=avoid_overwriting)




# -*- coding: utf-8; -*-
"""
Copyright (c) 2018 Rolf Hempel, rolf6419@gmx.de

This file is part of the PlanetarySystemStacker tool (PSS).
https://github.com/Rolf-Hempel/PlanetarySystemStacker

PSS is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PSS.  If not, see <http://www.gnu.org/licenses/>.

Part of this module (in class "AlignmentPointEditor" was copied from
https://stackoverflow.com/questions/35508711/how-to-enable-pan-and-zoom-in-a-qgraphicsview

"""
from glob import glob
from sys import argv, exit
from time import time

from PyQt5 import QtCore, QtWidgets
from numpy import uint8, uint16, int32

from align_frames import AlignFrames
from alignment_point_editor import SelectionRectangleGraphicsItem
from configuration import Configuration
from exceptions import InternalError, NotSupportedError, Error
from frame_viewer import FrameViewer
from frames import Frames
from rank_frames import RankFrames
from rectangular_patch_editor_gui import Ui_rectangular_patch_editor


class GraphicsScene(QtWidgets.QGraphicsScene):
    """
    Handle mouse events on the graphics scene depicting the image with a rectangular patch
    superimposed. The left mouse button can be used to create the patch.

    """

    def __init__(self, photo_editor, parent=None):
        """
        Initialize the scene object.

        :param photo_editor: the object in which the scene is defined
        :param parent: The parent class
        """

        QtWidgets.QGraphicsScene.__init__(self, parent)
        self.photo_editor = photo_editor
        self.left_button_pressed = False
        self.right_button_pressed = False

        self.remember_rp = None

    def mousePressEvent(self, event):
        """
        A mouse button is pressed.

        :param event: event object
        :return: -
        """

        # The following actions are not performed in drag-and-zoom mode. The switch between both
        # modes is handled in the higher-level object "photo_viewer".
        if not self.photo_editor.drag_mode and self.photo_editor.hasPhoto():

            # If a rectangular patch had been drawn before, remove it from the scene.
            if self.remember_rp:
                self.removeItem(self.remember_rp)
                self.remember_rp = None

            pos = event.lastScenePos()
            x = int(pos.x())
            y = int(pos.y())

            # The left button is pressed.
            if event.button() == QtCore.Qt.LeftButton:
                self.left_button_pressed = True

                # Remember the location.
                self.left_y_start = y
                self.left_x_start = x


    def mouseReleaseEvent(self, event):
        """
        A mouse button is released.

        :param event: event object
        :return: -
        """

        if not self.photo_editor.drag_mode and self.photo_editor.hasPhoto():

            # The left button is released.
            if event.button() == QtCore.Qt.LeftButton:
                self.left_button_pressed = False

                # If a rectangular patch has been defined, store its coordinates.
                if self.remember_rp:
                    self.photo_editor.set_selection_rectangle(self.remember_rp.y_low,
                                                              self.remember_rp.y_high,
                                                              self.remember_rp.x_low,
                                                              self.remember_rp.x_high)
                # If the rectangle is void, set all coordinates to zero.
                else:
                    self.photo_editor.set_selection_rectangle(0, 0, 0, 0)



    def mouseMoveEvent(self, event):
        """
        The mouse is moved while being pressed.

        :param event: event object
        :return: -
        """

        pos = event.lastScenePos()
        x = int(pos.x())
        y = int(pos.y())

        if not self.photo_editor.drag_mode and self.photo_editor.hasPhoto():

            # The mouse is moved with the left button pressed.
            if self.left_button_pressed:

                # Compute the new rectangular patch.
                new_sp = SelectionRectangleGraphicsItem(self.left_y_start, self.left_x_start, y, x)

                # If the rectangle was drawn for a previous location, replace it with the new one.
                if self.remember_rp is not None:
                    self.removeItem(self.remember_rp)
                self.addItem(new_sp)
                self.remember_rp = new_sp

            self.update()


class RectangularPatchEditor(FrameViewer):
    """
    This widget implements an editor for handling a rectangular patch superimposed onto an image.
    It supports two modes:
    - In "drag mode" the mouse can be used for panning, and the scroll wheel for zooming.
    - In "alignment point mode" the mouse is used to create/remove APs, to move them or to change
      their sizes.
    The "cntrl" key is used to switch between the two modes.
    """

    def __init__(self, image):
        super(RectangularPatchEditor, self).__init__()

        self.image = image

        # Initialize the rectangular patch.
        self.y_low = None
        self.y_high = None
        self.x_low = None
        self.x_high = None

        self.setPhoto(self.image)

    def initialize_scene(self):
        """
        Initialize the scene. This object handles mouse events if not in drag mode. In derived
        viewer classes this method is replaced with the instantiation of a custom version of the
        graphics scene.

        :return:
        """

        self._scene = GraphicsScene(self, self)
        self._scene.addItem(self._photo)
        self.setScene(self._scene)

    def set_selection_rectangle(self, y_low, y_high, x_low, x_high):
        """
        This method is invoked from the underlying scene object when the left mouse button is
        released, and thus a rectangular patch is defined.

        :param y_low: Lower y bound of patch
        :param y_high: Upper y bound of patch
        :param x_low: Lower x bound of patch
        :param x_high: Upper x bound of patch
        :return: -
        """

        self.y_low = max(0, y_low)
        self.y_high = min(self.shape_y, y_high)
        self.x_low = max(0, x_low)
        self.x_high = min(self.shape_x, x_high)

    def selection_rectangle(self):
        """
        Return the coordinate bounds of the selected rectangle.

        :return: Tuple with 4 ints: (lower y bound, upper y bound, lower x bound, upper x bound)
        """

        if self.y_low is not None:
            return (self.y_low, self.y_high, self.x_low, self.x_high)
        else:
            return (0, 0, 0, 0)


class RectangularPatchEditorWidget(QtWidgets.QFrame, Ui_rectangular_patch_editor):
    """
    This widget implements a rectangular patch editor, to be used for the selection of the
    stabilization patch and the ROI rectangle.
    """

    def __init__(self, parent_gui, frame, message, signal_finished):
        """
        Initialization of the widget.

        :param parent_gui: Parent GUI object
        :param frame: Background image on which the patch is superimposed. Usually, the mean frame
                      is used for this purpose.
        :param message: Message to tell the user what to do.
        :param signal_finished: Qt signal with signature (int, int, int, int) sending the
                                coordinate bounds (y_low, y_high, x_low, x_high) of the patch
                                selected, or (0, 0, 0, 0) if unsuccessful.
        """

        super(RectangularPatchEditorWidget, self).__init__(parent_gui)
        self.setupUi(self)

        self.parent_gui = parent_gui

        # Convert the frame into uint8 format. If the frame type is uint16, values
        # correspond to 16bit resolution.
        if frame.dtype == uint16 or frame.dtype == int32:
            self.frame = (frame / 256).astype(uint8)
        elif frame.dtype == uint8:
            self.frame = frame
        else:
            raise NotSupportedError("Attempt to set a photo in frame viewer with type neither"
                                    " uint8 nor uint16 not int32")

        self.message = message
        self.signal_finished = signal_finished

        self.viewer = RectangularPatchEditor(self.frame)
        self.verticalLayout.insertWidget(0, self.viewer)

        self.messageLabel.setText(self.message)
        self.messageLabel.setStyleSheet('color: red')
        self.buttonBox.accepted.connect(self.done)
        self.buttonBox.rejected.connect(self.reject)
        self.shape_y = None
        self.shape_x = None
        self.y_low = None
        self.y_high = None
        self.x_low = None
        self.x_high = None

    def done(self):
        """
        On exit from the rectangular patch editor, look up the coordinate bounds of the rectangular
        patch.

        :return: -
        """

        if self.viewer.selection_rectangle() is not None:
            self.y_low, self.y_high, self.x_low, self.x_high = self.viewer.selection_rectangle()

        # If the patch was selected successfully, send the patch bounds to the workflow thread.
        # If it was not successful, send (0, 0, 0, 0).
        if self.parent_gui is not None:
            if self.y_low is not None:
                self.signal_finished.emit(self.y_low, self.y_high,
                                                         self.x_low, self.x_high)
            # If the patch is not valid, do frame alignment in automatic mode. This is done if
            # all bounds are zero.
            else:
                self.signal_finished.emit(0, 0, 0, 0)

        # Close the Window.
        self.close()

    def reject(self):
        """
        If the "cancel" button is pressed, the coordinates are not stored.

        :return: -
        """

        # No rectangle was selected. Continue the workflow in automatic mode.
        self.signal_finished.emit(0, 0, 0, 0)
        self.close()


if __name__ == '__main__':
    # Images can either be extracted from a video file or a batch of single photographs. Select
    # the example for the test run.
    type = 'video'
    if type == 'image':
        names = glob('Images/2012*.tif')
        # names = glob.glob('Images/Moon_Tile-031*ap85_8b.tif')
        # names = glob.glob('Images/Example-3*.jpg')
    else:
        names = 'Videos/another_short_video.avi'
    print(names)

    # Get configuration parameters.
    configuration = Configuration()
    configuration.initialize_configuration()
    try:
        frames = Frames(configuration, names, type=type)
        print("Number of images read: " + str(frames.number))
        print("Image shape: " + str(frames.shape))
    except Error as e:
        print("Error: " + e.message)
        exit()

    # Rank the frames by their overall local contrast.
    rank_frames = RankFrames(frames, configuration)
    start = time()
    rank_frames.frame_score()
    end = time()
    print('Elapsed time in ranking images: {}'.format(end - start))
    print("Index of maximum: " + str(rank_frames.frame_ranks_max_index))
    print("Frame scores: " + str(rank_frames.frame_ranks))
    print("Frame scores (sorted): " + str(
        [rank_frames.frame_ranks[i] for i in rank_frames.quality_sorted_indices]))
    print("Sorted index list: " + str(rank_frames.quality_sorted_indices))

    # Initialize the frame alignment object.
    align_frames = AlignFrames(frames, rank_frames, configuration)

    if configuration.align_frames_mode == "Surface":
        start = time()
        # Select the local rectangular patch in the image where the L gradient is highest in both x
        # and y direction. The scale factor specifies how much smaller the patch is compared to the
        # whole image frame.
        (y_low_opt, y_high_opt, x_low_opt, x_high_opt) = align_frames.compute_alignment_rect(
            configuration.align_frames_rectangle_scale_factor)
        end = time()
        print('Elapsed time in computing optimal alignment rectangle: {}'.format(end - start))
        print("optimal alignment rectangle, x_low: " + str(x_low_opt) + ", x_high: " + str(
            x_high_opt) + ", y_low: " + str(y_low_opt) + ", y_high: " + str(y_high_opt))
        reference_frame_with_alignment_points = frames.frames_mono(
            rank_frames.frame_ranks_max_index).copy()
        reference_frame_with_alignment_points[y_low_opt,
        x_low_opt:x_high_opt] = reference_frame_with_alignment_points[y_high_opt - 1,
                                x_low_opt:x_high_opt] = 255
        reference_frame_with_alignment_points[y_low_opt:y_high_opt,
        x_low_opt] = reference_frame_with_alignment_points[y_low_opt:y_high_opt,
                     x_high_opt - 1] = 255
        # plt.imshow(reference_frame_with_alignment_points, cmap='Greys_r')
        # plt.show()

    # Align all frames globally relative to the frame with the highest score.
    start = time()
    try:
        align_frames.align_frames()
    except NotSupportedError as e:
        print("Error: " + e.message)
        exit()
    except InternalError as e:
        print("Warning: " + e.message)
        for index, frame_number in enumerate(align_frames.failed_index_list):
            print("Shift computation failed for frame " + str(
                align_frames.failed_index_list[index]) + ", minima list: " + str(
                align_frames.dev_r_list[index]))
    end = time()
    print('Elapsed time in aligning all frames: {}'.format(end - start))
    print("Frame shifts: " + str(align_frames.frame_shifts))
    print("Intersection: " + str(align_frames.intersection_shape))

    start = time()
    # Compute the reference frame by averaging the best frames.
    average = align_frames.average_frame()
    border = configuration.align_frames_search_width
    # border = 100

    app = QtWidgets.QApplication(argv)
    window = RectangularPatchEditorWidget(None, average[border:-border, border:-border],
            "With 'ctrl' and the left mouse button pressed, draw a rectangular patch "
            "to be used for frame alignment. Or just press 'OK / Cancel' (automatic selection).",
            None)
    window.setMinimumSize(800, 600)
    window.showMaximized()
    app.exec_()

    print("Rectangle selected, y_low: " + str(border+window.y_low) + ", y_high: " +
           str(border+window.y_high) + ", x_low: " + str(border+window.x_low) + ", x_high: "
           + str(border+window.x_high))

    exit()




# -*- coding: utf-8; -*-
"""
Copyright (c) 2018 Rolf Hempel, rolf6419@gmx.de

This file is part of the PlanetarySystemStacker tool (PSS).
https://github.com/Rolf-Hempel/PlanetarySystemStacker

PSS is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PSS.  If not, see <http://www.gnu.org/licenses/>.

"""

from glob import glob
from os import path, remove
from pathlib import Path
from time import time

import numpy as np
from cv2 import imread, VideoCapture, CAP_PROP_FRAME_COUNT, cvtColor, COLOR_BGR2GRAY, \
    COLOR_BGR2RGB, GaussianBlur, Laplacian, CV_32F, COLOR_RGB2BGR, imwrite, convertScaleAbs
from scipy import misc

from configuration import Configuration
from exceptions import TypeError, ShapeError, ArgumentError, WrongOrderingError, Error
# from frames import Frames


class FramesOld(object):
    """
        This object stores the image data of all frames. Four versions of the original frames are
        used throughout the data processing workflow. They are (re-)used in the folliwing phases:
        1. Original (color) frames, type: uint8 / uint16
            - Frame stacking ("stack_frames.stack_frames")
        2. Monochrome version of 1., type: uint8
            - Computing the average frame (only average frame subset, "align_frames.average_frame")
        3. Gaussian blur added to 2., type: type: uint8
            - Aligning all frames ("align_frames.align_frames")
            - Frame stacking ("stack_frames.stack_frames")
        4. Down-sampled Laplacian of 3., type: uint8
            - Overall image ranking ("rank_frames.frame_score")
            - Ranking frames at alignment points("alignment_points.compute_frame_qualities")

        Currently all versions of all frames are stored during the entire workflow. As an
        alternative, a non-buffered version should be added.

    """

    def __init__(self, configuration, names, type='video', progress_signal=None):
        """
        Initialize the Frame object, and read all images. Images can be stored in a video file or
        as single images in a directory.

        :param configuration: Configuration object with parameters
        :param names: In case "video": name of the video file. In case "image": list of names for
                      all images.
        :param type: Either "video" or "image".
        :param progress_signal: Either None (no progress signalling), or a signal with the signature
                                (str, int) with the current activity (str) and the progress in
                                percent (int).
        """

        self.configuration = configuration
        self.progress_signal = progress_signal

        if type == 'image':
            # Use scipy.misc to read in image files.
            self.number = len(names)
            self.signal_step_size = max(int(self.number / 10), 1)

            self.frames_original = []
            for frame_index, path in enumerate(names):
                # After every "signal_step_size"th frame, send a progress signal to the main GUI.
                if self.progress_signal is not None and frame_index%self.signal_step_size == 0:
                    self.progress_signal.emit("Read all frames",
                                         int((frame_index / self.number) * 100.))
                # Read the next frame.
                frame = imread(path, -1)
                self.frames_original.append(frame)

            if self.progress_signal is not None:
                self.progress_signal.emit("Read all frames", 100)
            self.shape = self.frames_original[0].shape
            dt0 = self.frames_original[0].dtype

            # Test if all images have the same shape, color type and depth.
            # If not, raise an exception.
            for image in self.frames_original:
                if image.shape != self.shape:
                    raise ShapeError("Images have different size")
                elif len(self.shape) != len(image.shape):
                    raise ShapeError("Mixing grayscale and color images not supported")
                if image.dtype != dt0:
                    raise TypeError("Images have different type")

        elif type == 'video':
            # In case "video", use OpenCV to capture frames from video file. Revert the implicit
            # conversion from RGB to BGR in OpenCV input.
            cap = VideoCapture(names)
            self.number = int(cap.get(CAP_PROP_FRAME_COUNT))
            self.frames_original = []
            self.signal_step_size = max(int(self.number / 10), 1)
            for frame_index in range(self.number):
                # After every "signal_step_size"th frame, send a progress signal to the main GUI.
                if self.progress_signal is not None and frame_index%self.signal_step_size == 0:
                    self.progress_signal.emit("Read all frames", int((frame_index/self.number)*100.))
                # Read the next frame.
                ret, frame = cap.read()
                if ret:
                    self.frames_original.append(cvtColor(frame, COLOR_BGR2RGB))
                else:
                    raise IOError("Error in reading video frame")
            cap.release()
            if self.progress_signal is not None:
                self.progress_signal.emit("Read all frames", 100)
            self.shape = self.frames_original[0].shape
            dt0 = self.frames_original[0].dtype
        else:
            raise TypeError("Image type " + type + " not supported")

        # Monochrome images are stored as 2D arrays, color images as 3D.
        if len(self.shape) == 2:
            self.color = False
        elif len(self.shape) == 3:
            self.color = True
        else:
            raise ShapeError("Image shape not supported")

        # Set the depth value of all images to either 16 or 8 bits.
        if dt0 == 'uint16':
            self.depth = 16
        elif dt0 == 'uint8':
            self.depth = 8
        else:
            raise TypeError("Frame type " + str(dt0) + " not supported")

        # Initialize lists of monochrome frames (with and without Gaussian blur) and their
        # Laplacians.
        self.frames_monochrome = None
        self.frames_monochrome_blurred = None
        self.frames_monochrome_blurred_laplacian = None
        self.used_alignment_points = None

    def frames(self, index):
        """
        Look up the original frame object with a given index.

        :param index: Frame index
        :return: Frame with index "index".
        """

        # # Code for positioning at arbitrary index.
        # # Check for valid frame number.
        # if 0 <= index < self.number:
        #     # Set frame position
        #     cap.set(cv2.CAP_PROP_POS_FRAMES, index)

        if not 0<=index<self.number:
            raise ArgumentError("Frame index " + str(index) + " is out of bounds")
        # print ("Accessing frame " + str(index))
        return self.frames_original[index]

    def frames_mono(self, index):
        """
        Look up the monochrome version of the frame object with a given index.

        :param index: Frame index
        :return: Monochrome frame with index "index".
        """

        if not 0 <= index < self.number:
            raise ArgumentError("Frame index " + str(index) + " is out of bounds")
        if self.frames_monochrome is not None:
            # print("Accessing frame monochrome " + str(index))
            return self.frames_monochrome[index]
        else:
            raise WrongOrderingError("Attempt to look up a monochrome frame before computing it")

    def frames_mono_blurred(self, index):
        """
        Look up a Gaussian-blurred frame object with a given index.

        :param index: Frame index
        :return: Gaussian-blurred frame with index "index".
        """

        if not 0 <= index < self.number:
            raise ArgumentError("Frame index " + str(index) + " is out of bounds")
        if self.frames_monochrome_blurred is not None:
            # print("Accessing frame with Gaussian blur " + str(index))
            return self.frames_monochrome_blurred[index]
        else:
            raise WrongOrderingError("Attempt to look up a Gaussian-blurred frame version before"
                                     " computing it")

    def frames_mono_blurred_laplacian(self, index):
        """
        Look up a Laplacian-of-Gaussian of a frame object with a given index.

        :param index: Frame index
        :return: LoG of a frame with index "index".
        """

        if not 0 <= index < self.number:
            raise ArgumentError("Frame index " + str(index) + " is out of bounds")
        if self.frames_monochrome_blurred_laplacian is not None:
            # print("Accessing LoG number " + str(index))
            return self.frames_monochrome_blurred_laplacian[index]
        else:
            raise WrongOrderingError("Attempt to look up a LoG frame version before computing it")

    def reset_alignment_point_lists(self):
        """
        Every frame keeps a list with the alignment points for which this frame is among the
        sharpest ones (so it is used in stacking). Reset this list for all frames.

        :return: -
        """

        # For every frame initialize the list with used alignment points.
        self.used_alignment_points = [[] for frame_index in range(self.number)]

    def add_monochrome(self, color):
        """
        Create monochrome versions of all frames. Add a list of monochrome frames
        "self.frames_mono". If the original frames are monochrome, just point the monochrome frame
        list to the original images (no deep copy!). Also, add a blurred version of the frame list
        (using a Gaussian filter) "self.frames_mono_blurred", and the Laplacian of that image.

        :param color: Either "red" or "green", "blue", or "panchromatic"
        :return: -
        """

        if self.color:
            colors = ['red', 'green', 'blue', 'panchromatic']
            if not color in colors:
                raise ArgumentError("Invalid color selected for channel extraction")
            self.frames_monochrome = []
        elif self.depth == 8:
            self.frames_monochrome = self.frames_original

        self.frames_monochrome_blurred = []
        self.frames_monochrome_blurred_laplacian = []

        for frame_index, frame in enumerate(self.frames_original):
            # After every "signal_step_size"th frame, send a progress signal to the main GUI.
            if self.progress_signal is not None and frame_index % self.signal_step_size == 0:
                self.progress_signal.emit("Gaussians / Laplacians",
                                          int((frame_index / self.number) * 100.))

            # If frames are in color mode, or the depth is 16bit, produce a 8bit B/W version.
            if self.color or self.depth != 8:
                # If frames are in color mode, create a monochrome version with same depth.
                if self.color:
                    if color == 'panchromatic':
                        frame_mono = cvtColor(frame, COLOR_BGR2GRAY)
                    else:
                        frame_mono = frame[:, :, colors.index(color)]
                else:
                    frame_mono = frame

                # If depth is larger than 8bit, reduce the depth to 8bit.
                if self.depth != 8:
                    frame_mono = ((frame_mono) / 255.).astype(np.uint8)

                self.frames_monochrome.append(frame_mono)

            # Add a version of the frame with Gaussian blur added.
            frame_monochrome_blurred = GaussianBlur(self.frames_monochrome[frame_index], (
                self.configuration.frames_gauss_width, self.configuration.frames_gauss_width), 0)
            self.frames_monochrome_blurred.append(frame_monochrome_blurred)

            # Compute the scaling factor for "convertScaleAbs" depending on the data type of the
            # monochrome image. The Laplacian is 8bit. If the monochrome image is 16bit, values
            # have to be scaled down.

            # Add the Laplacian of the down-sampled blurred image.
            if self.configuration.rank_frames_method == "Laplace":
                self.frames_monochrome_blurred_laplacian.append(convertScaleAbs(Laplacian(
                    frame_monochrome_blurred[::self.configuration.align_frames_sampling_stride,
                    ::self.configuration.align_frames_sampling_stride], CV_32F),
                    alpha=1))

        if self.progress_signal is not None:
            self.progress_signal.emit("Gaussians / Laplacians", 100)

    def save_image(self, filename, image, color=False, avoid_overwriting=True):
        """
        Save an image to a file.

        :param filename: Name of the file where the image is to be written
        :param image: ndarray object containing the image data
        :param color: If True, a three channel RGB image is to be saved. Otherwise, monochrome.
        :param avoid_overwriting: If True, append a string to the input name if necessary so that
                                  it does not match any existing file. If False, overwrite
                                  an existing file.
        :return: -
        """

        if avoid_overwriting:
            # If a file or directory with the given name already exists, append the word "_file".
            if Path(filename).is_dir():
                while True:
                    filename += '_file'
                    if not Path(filename).exists():
                        break
                filename += '.jpg'
            # If it is a file, try to append "_copy.tiff" to its basename. If it still exists, repeat.
            elif Path(filename).is_file():
                suffix = Path(filename).suffix
                while True:
                    p = Path(filename)
                    filename = Path.joinpath(p.parents[0], p.stem + '_copy' + suffix)
                    if not Path(filename).exists():
                        break
            else:
                # If the file name is new and has no suffix, add ".tiff".
                suffix = Path(filename).suffix
                if not suffix:
                    filename += '.tiff'

        # Don't care if a file with the given name exists. Overwrite it if necessary.
        elif path.exists(filename):
            remove(filename)

        # Write the image to the file. Before writing, convert the internal RGB representation into
        # the BGR representation assumed by OpenCV.
        if color:
            imwrite(str(filename), cvtColor(image, COLOR_RGB2BGR))
        else:
            imwrite(str(filename), image)


if __name__ == "__main__":

    # Images can either be extracted from a video file or a batch of single photographs. Select
    # the example for the test run.
    type = 'image'
    if type == 'image':
        # names = glob('Images/2012_*.tif')
        # names = glob('D:\SW-Development\Python\PlanetarySystemStacker\Examples\Moon_2011-04-10\South\*.TIF')
        names = glob('D:\SW-Development\Python\PlanetarySystemStacker\Examples\Moon_2019-01-20\Images\*.TIF')
    else:
        names = 'Videos/short_video.avi'

    # Get configuration parameters.
    configuration = Configuration()
    configuration.initialize_configuration()

    start = time()
    try:
        frames_old = FramesOld(configuration, names, type=type)
        print("Frames old: Number of images read: " + str(frames_old.number))
        print("Image shape: " + str(frames_old.shape))
    except Error as e:
        print("Error: " + e.message)
        exit()
    end = time()
    print("Read " + str(frames_old.number) + " frames in " + str(end-start) + " seconds.")

    # Create monochrome versions of all frames. If the original frames are monochrome, just point
    # the monochrome frame list to the original images (no deep copy!).
    try:
        frames_old.add_monochrome(configuration.frames_mono_channel)
    except ArgumentError as e:
        print("Error: " + e.message)
        exit()

    # try:
    #     frames = Frames(configuration, names, type=type,
    #                     buffer_original=False, buffer_monochrome=False, buffer_gaussian=False,
    #                     buffer_laplacian=False)
    #     print("\nFrames: Number of images: " + str(frames.number))
    #     print("Image shape: " + str(frames.shape))
    # except Error as e:
    #     print("Error: " + e.message)
    #     exit()
    #
    # indices = [1, 0, 2, 2, 8]
    #
    # for index in indices:
    #     frame_old = frames_old.frames_mono_blurred_laplacian(index)
    #     frame = frames.frames_mono_blurred_laplacian(index)
    #     difference = np.max(np.abs(frame - frame_old))
    #     print ("Error for frame " + str(index) + ": " + str(difference))



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'shift_distribution_viewer_gui.ui'
#
# Created by: PyQt5 UI code generator 5.13.0
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_shift_distribution_viewer(object):
    def setupUi(self, shift_distribution_viewer):
        shift_distribution_viewer.setObjectName("shift_distribution_viewer")
        shift_distribution_viewer.resize(900, 530)
        self.verticalLayout = QtWidgets.QVBoxLayout(shift_distribution_viewer)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.failedShiftsLabel = QtWidgets.QLabel(shift_distribution_viewer)
        font = QtGui.QFont()
        font.setPointSize(8)
        self.failedShiftsLabel.setFont(font)
        self.failedShiftsLabel.setObjectName("failedShiftsLabel")
        self.horizontalLayout.addWidget(self.failedShiftsLabel)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.buttonBox = QtWidgets.QDialogButtonBox(shift_distribution_viewer)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.horizontalLayout.addWidget(self.buttonBox)
        self.horizontalLayout.setStretch(1, 1)
        self.verticalLayout.addLayout(self.horizontalLayout)

        self.retranslateUi(shift_distribution_viewer)
        QtCore.QMetaObject.connectSlotsByName(shift_distribution_viewer)

    def retranslateUi(self, shift_distribution_viewer):
        _translate = QtCore.QCoreApplication.translate
        shift_distribution_viewer.setWindowTitle(_translate("shift_distribution_viewer", "Form"))
        self.failedShiftsLabel.setText(_translate("shift_distribution_viewer", "Failed shift measurements (percent):"))



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'frame_selector_gui.ui'
#
# Created by: PyQt5 UI code generator 5.14.2
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_frame_selector(object):
    def setupUi(self, frame_selector):
        frame_selector.setObjectName("frame_selector")
        frame_selector.resize(900, 630)
        frame_selector.setLocale(QtCore.QLocale(QtCore.QLocale.English, QtCore.QLocale.UnitedStates))
        frame_selector.setFrameShape(QtWidgets.QFrame.Panel)
        frame_selector.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.gridLayout = QtWidgets.QGridLayout(frame_selector)
        self.gridLayout.setObjectName("gridLayout")
        self.buttonBox = QtWidgets.QDialogButtonBox(frame_selector)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.gridLayout.addWidget(self.buttonBox, 2, 3, 1, 2)
        self.pushButton_play = QtWidgets.QPushButton(frame_selector)
        self.pushButton_play.setObjectName("pushButton_play")
        self.gridLayout.addWidget(self.pushButton_play, 2, 2, 1, 1)
        self.pushButton_stop = QtWidgets.QPushButton(frame_selector)
        self.pushButton_stop.setObjectName("pushButton_stop")
        self.gridLayout.addWidget(self.pushButton_stop, 2, 1, 1, 1)
        self.slider_frames = QtWidgets.QSlider(frame_selector)
        self.slider_frames.setMaximum(1000)
        self.slider_frames.setPageStep(1)
        self.slider_frames.setOrientation(QtCore.Qt.Horizontal)
        self.slider_frames.setObjectName("slider_frames")
        self.gridLayout.addWidget(self.slider_frames, 2, 0, 1, 1)
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.groupBox = QtWidgets.QGroupBox(frame_selector)
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.groupBox)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.listWidget = QtWidgets.QListWidget(self.groupBox)
        self.listWidget.setObjectName("listWidget")
        self.verticalLayout_2.addWidget(self.listWidget)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.addButton = QtWidgets.QPushButton(self.groupBox)
        self.addButton.setObjectName("addButton")
        self.horizontalLayout.addWidget(self.addButton)
        self.removeButton = QtWidgets.QPushButton(self.groupBox)
        self.removeButton.setObjectName("removeButton")
        self.horizontalLayout.addWidget(self.removeButton)
        self.verticalLayout_2.addLayout(self.horizontalLayout)
        self.verticalLayout.addWidget(self.groupBox)
        self.gridLayout.addLayout(self.verticalLayout, 0, 4, 1, 1)
        self.GroupBox_frame_sorting = QtWidgets.QGroupBox(frame_selector)
        self.GroupBox_frame_sorting.setObjectName("GroupBox_frame_sorting")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.GroupBox_frame_sorting)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.radioButton_quality = QtWidgets.QRadioButton(self.GroupBox_frame_sorting)
        self.radioButton_quality.setObjectName("radioButton_quality")
        self.horizontalLayout_2.addWidget(self.radioButton_quality)
        self.radioButton_chronological = QtWidgets.QRadioButton(self.GroupBox_frame_sorting)
        self.radioButton_chronological.setObjectName("radioButton_chronological")
        self.horizontalLayout_2.addWidget(self.radioButton_chronological)
        self.gridLayout.addWidget(self.GroupBox_frame_sorting, 1, 4, 1, 1)
        self.gridLayout.setColumnStretch(0, 7)
        self.gridLayout.setColumnStretch(4, 3)

        self.retranslateUi(frame_selector)
        QtCore.QMetaObject.connectSlotsByName(frame_selector)

    def retranslateUi(self, frame_selector):
        _translate = QtCore.QCoreApplication.translate
        frame_selector.setWindowTitle(_translate("frame_selector", "Frame"))
        self.buttonBox.setToolTip(_translate("frame_selector", "Exit the viewer. Press \'OK\' to save the selection, or \'cancel\' to discard changes."))
        self.pushButton_play.setToolTip(_translate("frame_selector", "Start a frame display video."))
        self.pushButton_play.setText(_translate("frame_selector", "Play"))
        self.pushButton_stop.setToolTip(_translate("frame_selector", "Stop the frame display video."))
        self.pushButton_stop.setText(_translate("frame_selector", "Stop"))
        self.slider_frames.setToolTip(_translate("frame_selector", "Use the slider to select the frame to be displayed. As an alternative,\n"
"you can select the frame in the \'Selecct / deselect frames\' list."))
        self.groupBox.setTitle(_translate("frame_selector", "Select / deselect frames"))
        self.listWidget.setToolTip(_translate("frame_selector", "Left-click to mark frame indices / index ranges. Include / exclude selected frames via the context menu, the plus / minus buttons below, or by pressing + / - on the keyboard."))
        self.addButton.setToolTip(_translate("frame_selector", "Use selected frame(s) for stacking."))
        self.addButton.setText(_translate("frame_selector", "+"))
        self.removeButton.setToolTip(_translate("frame_selector", "Don\'t use selected frame(s) for stacking."))
        self.removeButton.setText(_translate("frame_selector", "-"))
        self.GroupBox_frame_sorting.setTitle(_translate("frame_selector", "Frame sorting"))
        self.radioButton_quality.setToolTip(_translate("frame_selector", "Frames are ordered by their overall sharpness."))
        self.radioButton_quality.setText(_translate("frame_selector", "By quality"))
        self.radioButton_chronological.setToolTip(_translate("frame_selector", "Frames are ordered chronologically."))
        self.radioButton_chronological.setText(_translate("frame_selector", "Chronological"))
